import dataclasses
from typing import Dict

import torch
import torch.nn as nn
from tinyspace import Space, sample_from_space

from multibeast.builder import __PolicyNet__
from multibeast.distributions import get_action_params


@dataclasses.dataclass
class ExampleLearnerState:
    r"""See impala.ImpalaLearnerState."""
    pass


class ExampleModel(nn.Module):
    def __init__(
        self,
        observation_space: Space,
        action_space: Space,
        action_dist_params: Dict = None,
        policy_params: Dict = None,
        feature_extractor: nn.Module = None,
    ):
        super().__init__()
        self.observation_space = observation_space
        self.action_space = action_space

        self.num_actions, self.max_action, self.action_dist_params = get_action_params(
            action_space,
            action_dist_params,
        )

        fc_out_features = 256

        self.feature_extractor = feature_extractor
        if feature_extractor:
            with torch.no_grad():
                x = sample_from_space(self.observation_space, batch_size=(1, 1), to_torch_tensor=True)
                feature_dim = tuple(self.feature_extractor(x).squeeze(0).shape)[0]
            self.fc = nn.Sequential(nn.ReLU(inplace=True), nn.Linear(feature_dim, fc_out_features), nn.ReLU(inplace=True))
        else:
            feature_dim = observation_space["shape"][0]
            self.fc = nn.Sequential(nn.Linear(feature_dim, fc_out_features), nn.ReLU(inplace=True))

        core_output_size = fc_out_features

        self.policy = __PolicyNet__.build(policy_params, core_output_size, self.num_actions, self.action_dist_params)

    def initial_state(self, batch_size=1):
        return tuple()

    def forward(self, inputs, core_state=None, deterministic: bool = False):
        prev_action = inputs["prev_action"]
        reward = inputs["reward"]
        x = inputs["state"]
        T, B = reward.shape

        # see impala.ImpalaNet for example on using core_state for lstm
        _ = self.policy(x, prev_action)

        output = dict()
        return output, core_state


class ExampleAgent:
    @staticmethod
    def create_agent(FLAGS, observation_space, action_space):
        r"""This function creates a `nn.Module` model and `LearnerState` for optimizing an agent.

        returns: `nn.Module`, `LearnerState`
        """
        raise NotImplementedError

    @staticmethod
    def step_optimizer(FLAGS, learner_state, stats):
        r"""This function should call `optimizer.step()` to update model parameters.

        returns: `None`
        """
        raise NotImplementedError

    @staticmethod
    def compute_gradients(FLAGS, data, learner_state, stats):
        r"""This function should compute a loss and call `loss.backward()`.

        returns: `None`
        """
        raise NotImplementedError

    @staticmethod
    def create_stats(FLAGS):
        r"""This function return a list of strings representing keys of statistics to track.

        returns: `List[str]`
        """
        raise NotImplementedError



# Copyright (c) Facebook, Inc. and its affiliates.
import moolib
import argparse
import time

# TODO: Can moolib choose a port for us?
DEFAULT_PORT = 4431

parser = argparse.ArgumentParser(description="A script to run a moolib broker")

parser.add_argument(
    "address",
    nargs="?",
    default="0.0.0.0:%i" % DEFAULT_PORT,
    type=str,
    metavar="addr:port",
    help="Broker server address to listen on.",
)


def main():
    FLAGS = parser.parse_args()

    broker_rpc = moolib.Rpc()
    broker_rpc.set_name("broker")
    broker = moolib.Broker(broker_rpc)
    broker_rpc.listen(FLAGS.address)

    print("Broker listening at %s" % FLAGS.address)

    try:
        while True:
            broker.update()
            time.sleep(0.25)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



import copy
import dataclasses
import logging

import torch
import moolib


@dataclasses.dataclass
class StatMean:
    value: float = 0
    n: int = 0

    def result(self):
        if self.n == 0:
            return None
        return self.value / self.n

    def __sub__(self, other):
        assert isinstance(other, StatMean)
        return StatMean(self.value - other.value, self.n - other.n)

    def __iadd__(self, other):
        if isinstance(other, StatMean):
            self.value += other.value
            self.n += other.n
        else:
            self.value += other
            self.n += 1
        return self

    def reset(self):
        self.value = 0
        self.n = 0

    def __repr__(self):
        return repr(self.result())


@dataclasses.dataclass
class StatSum:
    value: float = 0

    def result(self):
        return self.value

    def __sub__(self, other):
        assert isinstance(other, StatSum)
        return StatSum(self.value - other.value)

    def __iadd__(self, other):
        if isinstance(other, StatSum):
            self.value += other.value
        else:
            self.value += other
        return self

    def reset(self):
        pass

    def __repr__(self):
        return repr(self.result())


class GlobalStatsAccumulator:
    """Class for global accumulation state. add_stats gets reduced."""

    def __init__(self, rpc_group, global_stats):
        self.rpc_group = rpc_group
        self.global_stats = global_stats
        self.reduce_future = None
        self.queued_global_stats = None
        self.sent_global_stats = None
        self.prev_stats = None

    def add_stats(self, dst, src):
        for k, v in dst.items():
            v += src[k]
        return dst

    def enqueue_global_stats(self, stats):
        if self.queued_global_stats is None:
            self.queued_global_stats = copy.deepcopy(stats)
        else:
            # Sum pending data.
            self.add_stats(self.queued_global_stats, stats)

    def reduce(self, stats):
        if self.reduce_future is not None and self.reduce_future.done():
            if self.reduce_future.exception() is not None:
                logging.info(
                    "global stats accumulation error: %s",
                    self.reduce_future.exception(),
                )
                self.enqueue_global_stats(self.sent_global_stats)
            else:
                self.add_stats(self.global_stats, self.reduce_future.result())
            self.reduce_future = None

        stats_diff = stats
        if self.prev_stats is not None:
            stats_diff = {k: v - self.prev_stats[k] for k, v in stats.items()}

        self.enqueue_global_stats(stats_diff)
        self.prev_stats = copy.deepcopy(stats)

        if self.reduce_future is None:
            # Only reduce when not currently reducing.
            # Otherwise, we keep queued_global_stats for next time.
            self.sent_global_stats = self.queued_global_stats
            self.queued_global_stats = None
            # Additional copy to deal with potential partial reductions.
            self.reduce_future = self.rpc_group.all_reduce(
                "global stats", copy.deepcopy(self.sent_global_stats), self.add_stats
            )

    def reset(self):
        if self.prev_stats is not None:
            for _, v in self.prev_stats.items():
                v.reset()


def _mean_var_count_from_moments(mean, var, count, batch_mean, batch_var, batch_count):
    delta = batch_mean - mean
    tot_count = count + batch_count

    new_mean = mean + delta * batch_count / tot_count
    m_a = var * count
    m_b = batch_var * batch_count
    M2 = m_a + m_b + torch.square(delta) * count * batch_count / tot_count
    new_var = M2 / tot_count
    new_count = tot_count

    return new_mean, new_var, new_count


# From https://github.com/openai/baselines/blob/ea25b9e8b234e6ee1bca43083f8f3cf974143998/baselines/common/running_mean_std.py#L5
class RunningMeanStd(object):
    # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
    def __init__(self, epsilon=1e-4, shape=()):
        self.mean = torch.zeros(shape, dtype=torch.float64)
        self.var = torch.ones(shape, dtype=torch.float64)
        self.count = epsilon

    def update(self, x):
        batch_mean = torch.mean(x, axis=0)
        batch_var = torch.var(x, axis=0)
        batch_count = x.shape[0]
        self.mean, self.var, self.count = _mean_var_count_from_moments(
            self.mean, self.var, self.count, batch_mean, batch_var, batch_count
        )


class EnvBatchState:
    def __init__(self, flags, model):
        batch_size = flags.actor_batch_size
        device = flags.device
        self.batch_size = batch_size
        self.prev_action = torch.zeros(batch_size).long().to(device)
        self.future = None
        self.core_state = model.initial_state(batch_size=batch_size)
        self.core_state = tuple(s.to(device) for s in self.core_state)
        self.initial_core_state = self.core_state

        self.running_reward = torch.zeros(batch_size)
        self.step_count = torch.zeros(batch_size)

        self.discounting = flags.discounting
        self.weighted_returns = torch.zeros(batch_size)
        self.weighted_returns_rms = RunningMeanStd()

        self.time_batcher = moolib.Batcher(flags.unroll_length + 1, device)

    def update(self, env_outputs, action, stats):
        self.prev_action = action
        self.running_reward += env_outputs["reward"]
        self.weighted_returns *= self.discounting
        self.weighted_returns += env_outputs["reward"]
        self.weighted_returns_rms.update(self.weighted_returns)

        self.scaled_reward = env_outputs["reward"] / torch.sqrt(
            self.weighted_returns_rms.var + 1e-8
        )

        self.step_count += 1

        done = env_outputs["done"]

        episode_return = self.running_reward * done
        episode_step = self.step_count * done

        episodes_done = done.sum().item()
        if episodes_done > 0:
            stats["mean_episode_return"] += episode_return.sum().item() / episodes_done
            stats["mean_episode_step"] += episode_step.sum().item() / episodes_done
        stats["steps_done"] += done.numel()
        stats["episodes_done"] += episodes_done

        stats["running_reward"] += self.running_reward.mean().item()
        stats["running_step"] += self.step_count.mean().item()

        not_done = ~done

        self.running_reward *= not_done
        self.weighted_returns *= not_done
        self.step_count *= not_done



import torch

from multibeast.builder import __Distribution__, __FeatureExtractor__, __MakeEnv__, __PolicyNet__
from multibeast.distributions import get_action_params

from .mock_env import MockEnv


def test_MakeEnv_build():
    @__MakeEnv__.register()
    def make_env(env_flags):
        return MockEnv(**env_flags)

    action_space_cls = "discrete"
    create_env_fn = __MakeEnv__.build("make_env", dict(action_space_cls=action_space_cls))
    env = create_env_fn()
    assert env.action_space["cls"] == action_space_cls

    assert __MakeEnv__._undo_register(make_env.__name__)


def test_FeatureExtractor_build():
    @__FeatureExtractor__.register()
    class MockModule:
        def __init__(self, observation_space, action_space, n_layers=3):
            self.n_layers = n_layers

    env = MockEnv(action_space_cls="box")

    feature_extractor_flags = dict(cls="MockModule", n_layers=5)
    feature_extractor = __FeatureExtractor__.build(feature_extractor_flags, env.observation_space, env.action_space)
    assert feature_extractor.n_layers == 5

    assert __FeatureExtractor__._undo_register(MockModule.__name__)


def test_PolicyNet_build():
    @__PolicyNet__.register()
    class MockModule:
        def __init__(self, feature_dim, action_dim, action_dist_params, n_layers=3):
            self.n_layers = n_layers

            self.action_dist_cls = __Distribution__.build(action_dist_params)

        def action_dist(self, policy_logits):
            return self.action_dist_cls(logits=policy_logits)

    env = MockEnv(action_space_cls="discrete")

    action_dist_params = None
    num_actions, max_action, action_dist_params = get_action_params(env.action_space, action_dist_params)

    policy_net_flags = dict(cls="MockModule", n_layers=5)
    policy_net = __PolicyNet__.build(policy_net_flags, 128, num_actions, action_dist_params)
    assert policy_net.n_layers == 5
    policy_net.action_dist(torch.randn((20, 4)))

    assert __PolicyNet__._undo_register(MockModule.__name__)



import dataclasses
import logging
from typing import Optional

import torch
import torch.nn as nn
from moolib.examples import common

from multibeast.builder import __Agent__, __FeatureExtractor__

from . import vtrace
from .ppo_net import PPONet


@dataclasses.dataclass
class PPOLearnerState:
    model: torch.nn.Module
    optimizer: torch.optim.Optimizer
    scheduler: torch.optim.lr_scheduler.LambdaLR
    model_version: int = 0
    num_previous_leaders: int = 0
    train_time: float = 0
    last_checkpoint: float = 0
    last_checkpoint_history: float = 0
    global_stats: Optional[dict] = None

    def save(self):
        r = dataclasses.asdict(self)
        r["model"] = self.model.state_dict()
        r["optimizer"] = self.optimizer.state_dict()
        r["scheduler"] = self.scheduler.state_dict()
        return r

    def load(self, state):
        for k, v in state.items():
            if k not in ("model", "optimizer", "scheduler", "global_stats"):
                setattr(self, k, v)
        self.model_version = state["model_version"]
        self.model.load_state_dict(state["model"])
        self.optimizer.load_state_dict(state["optimizer"])
        self.scheduler.load_state_dict(state["scheduler"])

        for k, v in state["global_stats"].items():
            if k in self.global_stats:
                self.global_stats[k] = type(self.global_stats[k])(**v)


@__Agent__.register()
class PPO:
    def __new__(cls):
        raise TypeError("Static class cannot be instantiated")

    @staticmethod
    def _create_model(FLAGS, observation_space, action_space):
        if FLAGS.feature_extractor:
            feature_extractor = __FeatureExtractor__.build(
                FLAGS.feature_extractor,
                observation_space,
                action_space,
            )
        else:
            feature_extractor = None

        action_dist_params = FLAGS.get("action_dist_params", None)
        policy_params = FLAGS.get("policy_params", dict(cls="PolicyNet"))

        model = PPONet(
            observation_space,
            action_space,
            feature_extractor=feature_extractor,
            action_dist_params=action_dist_params,
            policy_params=policy_params,
            use_lstm=FLAGS.use_lstm,
        )

        logging.info(f"model: \n{model}")

        return model

    @staticmethod
    def _create_optimizer(FLAGS, model):
        return torch.optim.Adam(
            model.parameters(),
            lr=FLAGS.optimizer.learning_rate,
            betas=(FLAGS.optimizer.beta_1, FLAGS.optimizer.beta_2),
            eps=FLAGS.optimizer.epsilon,
        )

    @staticmethod
    def _create_scheduler(FLAGS, optimizer):
        factor = FLAGS.unroll_length * FLAGS.virtual_batch_size / FLAGS.total_steps
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: max(1 - epoch * factor, 0))
        return scheduler

    @staticmethod
    def create_agent(FLAGS, observation_space, action_space):
        model = PPO._create_model(FLAGS, observation_space, action_space)
        optimizer = PPO._create_optimizer(FLAGS, model)
        scheduler = PPO._create_scheduler(FLAGS, optimizer)
        learner_state = PPOLearnerState(model, optimizer, scheduler)
        return model, learner_state

    @staticmethod
    def step_optimizer(FLAGS, learner_state, stats):
        unclipped_grad_norm = nn.utils.clip_grad_norm_(learner_state.model.parameters(), FLAGS.grad_norm_clipping)
        learner_state.optimizer.step()
        learner_state.scheduler.step()
        learner_state.model_version += 1

        stats["unclipped_grad_norm"] += unclipped_grad_norm.item()
        stats["optimizer_steps"] += 1
        stats["model_version"] += 1
        return

    @staticmethod
    def compute_gradients(FLAGS, data, learner_state, stats):
        return vtrace.compute_gradients(FLAGS, data, learner_state, stats)

    @staticmethod
    def create_stats():
        stats_keys = [
            "entropy_loss",
            "pg_loss",
            "baseline_loss",
            "kl_loss",
            "total_loss",
            "unclipped_grad_norm",
        ]

        stats = {}
        for k in stats_keys:
            stats[k] = common.StatMean()
        return stats



import torch
import torch.nn.functional as F


def compute_baseline_loss(advantages):
    return 0.5 * torch.mean(advantages**2)


def compute_entropy_loss(logits):
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    entropy_per_timestep = torch.sum(-policy * log_policy, dim=-1)
    return -torch.mean(entropy_per_timestep)


def compute_policy_gradient_loss(logits, actions, advantages):
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    policy_gradient_loss_per_timestep = cross_entropy * advantages.detach()
    return torch.mean(policy_gradient_loss_per_timestep)


# from https://github.com/facebookresearch/minihack/blob/65fc16f0f321b00552ca37db8e5f850cbd369ae5/minihack/agent/polybeast/models/losses.py
# NOTE: those functions do not divide by the batch size


def compute_forward_dynamics_loss(pred_next_emb, next_emb):
    forward_dynamics_loss = torch.norm(pred_next_emb - next_emb, dim=2, p=2)
    return torch.sum(torch.mean(forward_dynamics_loss, dim=1))


def compute_inverse_dynamics_loss(pred_actions, true_actions):
    inverse_dynamics_loss = F.cross_entropy(
        torch.flatten(pred_actions, 0, 1),
        torch.flatten(true_actions, 0, 1),
        reduction="none",
    )
    inverse_dynamics_loss = inverse_dynamics_loss.view_as(true_actions)
    return torch.sum(torch.mean(inverse_dynamics_loss, dim=1))



import torch
import torch.distributions as D

from multibeast.distributions import SquashedDiagGaussian


# from https://github.com/denisyarats/pytorch_sac/blob/7aa74312d82be95a8ae42d62d82f47a0980553da/distribs.py#L38
class SquashedNormal(D.TransformedDistribution):
    def __init__(self, loc, scale, independent=False):
        self.loc = loc
        self.scale = scale
        self.independent = independent

        self.base_dist = D.Normal(loc, scale)
        if independent:
            self.base_dist = D.Independent(self.base_dist, 1)
        transforms = [D.TanhTransform(cache_size=1)]
        super().__init__(self.base_dist, transforms)

    @property
    def mean(self):
        mu = self.loc
        for tr in self.transforms:
            mu = tr(mu)
        return mu


def test_SquashedDiagGaussian():
    T, B = 4, 2
    num_actions = 10
    loc = torch.randn(T, B, num_actions)
    scale = torch.randn(T, B, num_actions).exp()

    n = 2

    d1 = SquashedDiagGaussian(loc, scale)
    a1 = [d1.rsample() for _ in range(n)]
    lp1 = [d1.log_prob(a) for a in a1]

    for a in a1:
        assert a.shape == (T, B, num_actions)
    for lp in lp1:
        assert lp.shape == (T, B)

    # verify that separating time & batch yields the same result as when combining them
    d1_flatten = SquashedDiagGaussian(loc.view(T * B, -1), scale.view(T * B, -1))
    for a, lp in zip(a1, lp1):
        lp_flatten = d1_flatten.log_prob(a.view(T * B, -1))
        assert lp_flatten.shape == (T * B,)
        assert torch.all(torch.isclose(lp, lp_flatten.view(T, B)))

    # # This is unstable, and will result in nans near the boundary b/c it does not clip
    # d2 = SquashedNormal(loc.view(T * B, -1), scale.view(T * B, -1), independent=True)
    # a2 = [d1.rsample() for _ in range(n)]
    # for a, lp in zip(a1, lp1):
    #     lp_flatten = d2.log_prob(a.view(T * B, -1))
    #     assert lp_flatten.shape == (T * B,)
    #     lp_ = lp_flatten.view(T, B)
    #     print(lp, lp_)
    #     assert torch.all(torch.isclose(lp, lp_))



import asyncio
import multiprocessing as mp
from typing import Any

import cloudpickle
import torch
from tinyspace import collate_obs


class CloudpickleWrapper:
    """Uses cloudpickle to serialize contents (otherwise multiprocessing tries to use pickle).

    Args:
        var: the variable you wish to wrap for pickling with cloudpickle
    """

    def __init__(self, var: Any):
        self.var = var

    def __getstate__(self) -> Any:
        return cloudpickle.dumps(self.var)

    def __setstate__(self, var: Any) -> None:
        self.var = cloudpickle.loads(var)


def _worker(remote, parent_remote, env_fn_wrapper):
    parent_remote.close()
    env = env_fn_wrapper.var()
    try:
        while True:
            try:
                cmd, data = remote.recv()
                if cmd == "step":
                    observation, reward, done, info = env.step(data)
                    if done:
                        # # NOTE: EnvBatchState.time_batcher deals with this instead
                        # # save final observation where user can get it, then reset
                        # info["terminal_observation"] = observation
                        observation = env.reset()
                    remote.send((observation, reward, done, info))
                elif cmd == "seed":
                    remote.send(env.seed(data))
                elif cmd == "reset":
                    observation = env.reset()
                    remote.send((observation, 0.0, False, {}))
                elif cmd == "render":
                    remote.send(env.render(data))
                elif cmd == "close":
                    env.close()
                    remote.close()
                    break
                elif cmd == "get_spaces":
                    remote.send((env.observation_space, env.action_space))
                elif cmd == "env_method":
                    method = getattr(env, data[0])
                    remote.send(method(*data[1], **data[2]))
                elif cmd == "get_attr":
                    remote.send(getattr(env, data))
                elif cmd == "set_attr":
                    remote.send(setattr(env, data[0], data[1]))
                else:
                    remote.close()
                    raise NotImplementedError(f"`{cmd}` is not implemented in the worker")
            except EOFError:
                break
    except KeyboardInterrupt:
        remote.close()


class EnvStepperFuture:
    def __init__(self, remotes, observation_space, callback_on_result=None):
        self.remotes = remotes
        self.observation_space = observation_space
        self.callback_on_result = callback_on_result

    def result(self):
        remotes_batch = self.remotes
        results = [remote.recv() for remote in remotes_batch]
        obses, rews, dones, infos = zip(*results)

        env_outputs = {}
        env_outputs["state"] = collate_obs(obses, self.observation_space, to_torch_tensor=True)
        env_outputs["reward"] = torch.tensor(rews, dtype=torch.float)
        env_outputs["done"] = torch.tensor(dones, dtype=torch.bool)
        env_outputs["info"] = {k: torch.tensor([x[k] for x in infos]) for k in infos[0].keys()}

        if self.callback_on_result is not None:
            self.callback_on_result()

        return env_outputs


class EnvStepper:
    r"""Steps a batch of environments.

    This class is based on stable_baselines3.SubprocVecEnv, and supports running a pool of
    unbatched CPU-based environments.

    Uses multiprocessing.Pipe instead of multiprocessing.SharedMemory, to support environments
    with observation spaces that contain variable-sized inputs (ie. point clouds).

    Args:
        env_init_fn: Callable to create a single environment.
        num_processes: Number of environments to run in parallel.
    """

    def __init__(self, env_init_fn, num_processes, start_method="spawn"):
        self.num_processes = num_processes
        self.busy_buffer = set()
        self.reset_once = set()
        self.closed = False

        env_fns = [env_init_fn] * num_processes

        ctx = mp.get_context(start_method)
        self.remotes, self.work_remotes = zip(*[ctx.Pipe() for _ in range(num_processes)])
        self.processes = []
        for work_remote, remote, env_fn in zip(self.work_remotes, self.remotes, env_fns):
            args = (work_remote, remote, CloudpickleWrapper(env_fn))
            # daemon=True: if the main process crashes, we should not cause things to hang
            process = ctx.Process(target=_worker, args=args, daemon=True)
            process.start()
            self.processes.append(process)
            work_remote.close()

        self.remotes[0].send(("get_spaces", None))
        self.observation_space, self.action_space = self.remotes[0].recv()

    def step(self, batch_index: int, actions: torch.Tensor) -> "asyncio.Future":
        batch_size = actions.size(0)
        idx = int(batch_index * batch_size)
        remotes_batch = self.remotes[idx : idx + batch_size]

        if batch_index in self.busy_buffer:
            raise RuntimeError(f"EnvStepper: attempt to step buffer index {batch_index} twice concurrently")

        self.busy_buffer.add(batch_index)

        if batch_index not in self.reset_once:
            for remote in self.remotes:
                remote.send(("reset", None))

            self.reset_once.add(batch_index)
        else:
            actions = actions.detach().cpu().numpy()
            for remote, action in zip(remotes_batch, actions):
                remote.send(("step", action))

        def callback_on_result():
            self.busy_buffer.remove(batch_index)

        return EnvStepperFuture(remotes_batch, self.observation_space, callback_on_result)

    def close(self):
        if self.closed:
            return
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
        for remote in self.remotes:
            remote.send(("close", None))
        for process in self.processes:
            process.join()
        self.closed = True



# Copyright (c) Facebook, Inc. and its affiliates.
"""'Logging' utilities."""

import csv
import json
import logging
import os
import time


def log_to_file(_state={}, **fields):  # noqa: B006
    """Incrementally write logs.tsv into pwd."""
    if "writer" not in _state:
        path = "logs.tsv"  # Could infer FLAGS if we had them.

        writeheader = not os.path.exists(path)
        fieldnames = list(fields.keys())

        _state["file"] = open(path, "a", buffering=1)  # Line buffering.
        _state["writer"] = csv.DictWriter(_state["file"], fieldnames, delimiter="\t")
        if writeheader:
            _state["writer"].writeheader()
            logging.info("Writing logs to %s", path)
        else:
            logging.info("Appending logs to %s", path)

    writer = _state["writer"]
    if writer is not None:
        writer.writerow(fields)


def symlink_path(target, symlink):
    try:
        if os.path.islink(symlink):
            os.remove(symlink)
        if not os.path.exists(symlink):
            os.symlink(target, symlink)
            return True
    except OSError:
        # os.remove() or os.symlink() raced. Don't do anything.
        pass
    return False


def write_metadata(localdir, srcdir, **kwargs):
    """Write meta.json file with some information on our setup."""
    if not localdir:
        return

    metadata = {
        "env": os.environ.copy(),
        "date_start": time.strftime("%Y-%m-%d %H:%M:%S"),
    }
    metadata.update(kwargs)

    try:
        import git
    except ImportError:
        logging.warning(
            "Couldn't import gitpython module; install it with `pip install gitpython`."
        )
    else:
        try:
            repo = git.Repo(path=srcdir, search_parent_directories=True)
            metadata["git"] = {
                "commit": repo.commit().hexsha,
                "is_dirty": repo.is_dirty(),
                "path": repo.git_dir,
            }
            if not repo.head.is_detached:
                metadata["git"]["branch"] = repo.active_branch.name
        except git.InvalidGitRepositoryError:
            pass

    if "git" not in metadata:
        logging.warning("Couldn't determine git data.")

    symlink = os.path.join(localdir, "meta.json")
    filename = "%s.%s.%d" % (symlink, time.strftime("%Y%m%d-%H%M%S"), os.getpid())

    with open(filename, "w") as f:
        json.dump(metadata, f, indent=2, sort_keys=True)

    symlink_path(os.path.relpath(filename, start=localdir), symlink)



import pytest
import tinyspace
import torch

from multibeast.agents.impala import ImpalaNet

from ..mock_env import MockEnv


def _get_input(env, T, B):
    x = tinyspace.sample_from_space(env.observation_space, batch_size=(T, B), to_torch_tensor=True)
    prev_action = tinyspace.sample_from_space(env.action_space, batch_size=(T, B), to_torch_tensor=True)
    inputs = dict(
        prev_action=prev_action,
        reward=torch.randn((T, B)),
        state=x,
        done=torch.zeros((T, B), dtype=torch.bool),
    )
    return inputs


@pytest.mark.order(1)
def test_impala_forward_Categorical(use_lstm=False):
    env = MockEnv(obs_space_type="1d", action_space_cls="discrete")

    T, B = 20, 4
    inputs = _get_input(env, T=T, B=B)

    # test if default is set properly
    policy_params = dict(cls="PolicyNet")
    action_dist_params = dict(cls="Categorical")
    model = ImpalaNet(env.observation_space, env.action_space)
    model = ImpalaNet(
        env.observation_space,
        env.action_space,
        policy_params=policy_params,
        action_dist_params=action_dist_params,
        use_lstm=use_lstm,
    )

    initial_state = model.initial_state(batch_size=B)

    outputs, core_state = model(inputs, core_state=initial_state)
    if use_lstm:
        assert len(core_state) == 2

    assert outputs["policy_logits"].shape == (T, B, model.num_actions)
    assert outputs["action"].shape == (T, B)
    assert outputs["baseline"].shape == (T, B)

    # try creating distribution
    action_dist = model.policy.action_dist(outputs["policy_logits"])
    assert action_dist.log_prob(outputs["action"]).shape == (T, B)


@pytest.mark.order(2)
def test_impala_forward_Categorical_use_lstm():
    test_impala_forward_Categorical(use_lstm=True)


@pytest.mark.order(2)
def test_impala_forward_SquashedDiagGaussian():
    env = MockEnv(obs_space_type="1d", action_space_cls="box")

    T, B = 20, 4
    inputs = _get_input(env, T=T, B=B)

    policy_params = dict(cls="PolicyNet", learn_std=False)
    action_dist_params = dict(cls="SquashedDiagGaussian")
    model = ImpalaNet(
        env.observation_space,
        env.action_space,
        policy_params=policy_params,
        action_dist_params=action_dist_params,
    )
    outputs, core_state = model(inputs)

    loc = outputs["policy_logits"][0]
    scale = outputs["policy_logits"][1]
    assert loc.shape == (T, B, model.num_actions)
    assert scale.shape == (T, B, model.num_actions)
    assert outputs["action"].shape == (T, B, model.num_actions)
    assert outputs["baseline"].shape == (T, B)

    # try creating distribution
    action_dist = model.policy.action_dist(outputs["policy_logits"])
    assert action_dist.log_prob(outputs["action"]).shape == (T, B)


@pytest.mark.order(2)
def test_impala_forward_DiscretizedLogisticMixture():
    env = MockEnv(obs_space_type="1d", action_space_cls="box")

    T, B = 20, 4
    inputs = _get_input(env, T, B)

    action_dist_params = dict(cls="DiscretizedLogisticMixture", num_bins=256)
    policy_params = dict(cls="MixturePolicyNet", n_mixtures=5, const_var=False, hidden_dim=None)
    model = ImpalaNet(
        env.observation_space,
        env.action_space,
        action_dist_params=action_dist_params,
        policy_params=policy_params,
    )
    outputs, core_state = model(inputs)

    for x in outputs["policy_logits"]:
        assert x.shape == (T, B, model.num_actions, model.policy._n_mixtures)
    assert outputs["action"].shape == (T, B, model.num_actions)
    assert outputs["baseline"].shape == (T, B)

    # try creating distribution
    action_dist = model.policy.action_dist(outputs["policy_logits"])
    assert action_dist.log_prob(outputs["action"]).shape == (T, B, model.num_actions)



import numpy as np
import torch
import torch.distributions as D
import torch.nn as nn

from multibeast.builder import __Distribution__, __PolicyNet__


class ExamplePolicyNet(nn.Module):
    def __init__(
        self,
        core_output_size,
        num_actions,
        action_dist_params,
    ):
        self.action_dist_cls = __Distribution__.build(action_dist_params)
        self.logits = nn.Linear(core_output_size, num_actions)

    def action_dist(self, policy_logits) -> D.Distribution:
        action_dist = self.action_dist_cls(logits=policy_logits)
        return action_dist

    def forward(self, core_output, deterministic: bool = False):
        T, B, _ = core_output.shape
        x = core_output.view(T * B, -1)
        policy_logits = self.logits(x)
        action_dist = self.action_dist(policy_logits)
        action = action_dist.rsample()  # preferred if has rsample(..) to pass gradients w/ reparameterization trick
        action = action.view(T, B, -1)
        policy_logits = policy_logits.view(T, B, -1)
        return policy_logits, action


@__PolicyNet__.register()
class PolicyNet(nn.Module):
    r"""This module supports up to two outputs (loc and scale) for some distribution parameterizing the policy.

    Refs:
    - https://github.com/ikostrikov/jaxrl/blob/8ac614b0c5202acb7bb62cdb1b082b00f257b08c/jaxrl/networks/policies.py#L14
    - https://github.com/DLR-RM/stable-baselines3/blob/ed308a71be24036744b5ad4af61b083e4fbdf83c/stable_baselines3/common/distributions.py#L115

    Args:
        learn_std: If None, then do not output stdev. If True, then std is learnable. If False, then
            std is kept constant to some initial value.
        state_dependent_std: If True, then use an additional layer to predict std given the state features.
        log_std_init: Initial value for the log standard deviation (using log std in fact to allow negative values)
    """

    def __init__(
        self,
        core_output_size,
        num_actions,
        action_dist_params,
        learn_std: bool = None,
        state_dependent_std: bool = False,
        log_std_init: float = -2.0,  # stable_baselines3 default, jaxrl uses 0.0
        log_std_min: float = -10.0,
        log_std_max: float = 2.0,
    ):
        super().__init__()
        self.action_dist_cls = __Distribution__.build(action_dist_params)

        self.learn_std = learn_std
        self.state_dependent_std = state_dependent_std
        self.log_std_init = log_std_init
        self.log_std_min = log_std_min
        self.log_std_max = log_std_max

        self.logits = nn.Linear(core_output_size, num_actions)

        if learn_std is not None:
            output_dim = num_actions

            if state_dependent_std:
                self._log_std = nn.Linear(core_output_size, output_dim)
            else:
                # -> diag covariance
                self._log_std = nn.Parameter(torch.ones(output_dim) * log_std_init, requires_grad=learn_std)

    def action_dist(self, policy_logits):
        if self.learn_std is None:
            return self.action_dist_cls(logits=policy_logits)
        else:
            assert isinstance(policy_logits, tuple)
            loc = policy_logits[0]
            scale = policy_logits[1]
            return self.action_dist_cls(loc, scale)

    def forward(self, core_output, deterministic: bool = False):
        T, B, _ = core_output.shape
        x = core_output.view(T * B, -1)
        policy_logits = self.logits(x)

        if self.learn_std is None:
            pass
        else:
            mu = policy_logits

            if self.state_dependent_std:
                scale = self._log_std(x)
            else:
                scale = self._log_std if self.training else self._log_std.detach()
                # NOTE: even though action_dist_cls may handle it, we expand scale to help moolib.Batcher
                scale = scale.view(1, -1).expand_as(mu)

            scale = torch.clamp(scale, self.log_std_min, self.log_std_max)
            scale = torch.exp(scale)
            policy_logits = (mu, scale)

        action_dist = self.action_dist(policy_logits)
        if deterministic:
            raise NotImplementedError
        else:
            # D.Categorical only defines sample()
            action = action_dist.rsample() if action_dist.has_rsample else action_dist.sample()

        if self.learn_std is None:
            action = action.view(T, B, -1)
            policy_logits = policy_logits.view(T, B, -1)
            return policy_logits, action
        else:
            action = action.view(T, B, -1)
            mu = mu.view(T, B, -1)
            scale = scale.view(T, B, -1)
            return (mu, scale), action


# from https://github.com/SudeepDasari/one_shot_transformers/blob/ecd43b0c182451b67219fdbc7d6a3cd912395f17/hem/models/inverse_module.py#L106
# They report "For most of our experiments, the model performed best when using two mixture
# components and learned constant variance parameters per action dimension", so `n_mixtures=2` and `const_var=True`.
# Also see https://github.com/ikostrikov/jaxrl/blob/8ac614b0c5202acb7bb62cdb1b082b00f257b08c/jaxrl/networks/policies.py#L47
@__PolicyNet__.register()
class MixturePolicyNet(nn.Module):
    def __init__(
        self,
        in_dim,
        out_dim,
        action_dist_params,
        n_mixtures: int = 3,
        const_var: bool = False,
        hidden_dim: int = 256,
    ):
        super().__init__()
        self.action_dist_cls = __Distribution__.build(action_dist_params)

        assert n_mixtures >= 1, "must predict at least one mixture!"
        self._n_mixtures, self._dist_size = n_mixtures, torch.Size((out_dim, n_mixtures))

        if hidden_dim:
            self._l = nn.Sequential(
                nn.Linear(in_dim, hidden_dim),
                nn.ReLU(inplace=True),
                nn.Linear(hidden_dim, hidden_dim),
                nn.ReLU(inplace=True),
            )
        else:
            self._l = None
            hidden_dim = in_dim

        self._mu = nn.Linear(hidden_dim, out_dim * n_mixtures)
        self._const_var = const_var

        if const_var:
            # independent of state, still optimized
            ln_scale = torch.randn(out_dim, dtype=torch.float32) / np.sqrt(out_dim)
            self.register_parameter("_ln_scale", nn.Parameter(ln_scale, requires_grad=True))
            # TODO: std per mixture?
        else:
            # state dependent
            self._ln_scale = nn.Linear(hidden_dim, out_dim * n_mixtures)
        self._logit_prob = nn.Linear(hidden_dim, out_dim * n_mixtures) if n_mixtures > 1 else None

    def action_dist(self, policy_logits_tuple):
        return self.action_dist_cls(*policy_logits_tuple)

    def forward(self, core_output, deterministic: bool = False):
        T, B, _ = core_output.shape

        x = core_output.view(T * B, -1)
        if self._l is not None:
            x = self._l(x)

        mu = self._mu(x).reshape((x.shape[:-1] + self._dist_size))
        if self._const_var:
            ln_scale = self._ln_scale if self.training else self._ln_scale.detach()
            ln_scale = ln_scale.reshape((1, -1, 1)).expand_as(mu)
        else:
            ln_scale = self._ln_scale(x).reshape((x.shape[:-1] + self._dist_size))

        logit_prob = (
            self._logit_prob(x).reshape((x.shape[:-1] + self._dist_size)) if self._n_mixtures > 1 else torch.ones_like(mu)
        )

        policy_logits_tuple = (mu, ln_scale, logit_prob)

        action_dist = self.action_dist(policy_logits_tuple)
        if deterministic:
            raise NotImplementedError
        else:
            action = action_dist.rsample()

        action = action.view(T, B, -1)
        policy_logits_tuple = tuple(x.view(T, B, -1, self._n_mixtures) for x in policy_logits_tuple)

        return policy_logits_tuple, action



import torch
import torch.nn.functional as F

from multibeast.builder import __Distribution__
from multibeast.distributions import get_action_params

from ..mock_env import MockEnv


def _compute_entropy_loss(logits):  # Categorical
    policy = F.softmax(logits, dim=-1)
    log_policy = F.log_softmax(logits, dim=-1)
    entropy_per_timestep = torch.sum(-policy * log_policy, dim=-1)
    return -torch.mean(entropy_per_timestep)


def _compute_policy_gradient_loss(logits, actions, advantages):  # Categorical
    cross_entropy = F.nll_loss(
        F.log_softmax(torch.flatten(logits, 0, 1), dim=-1),
        target=torch.flatten(actions, 0, 1),
        reduction="none",
    )
    cross_entropy = cross_entropy.view_as(advantages)
    policy_gradient_loss_per_timestep = cross_entropy * advantages.detach()
    return torch.mean(policy_gradient_loss_per_timestep)


def test_losses_categorical(test_entropy=True, test_kl=True, test_pg=True):
    env = MockEnv(obs_space_type="1d", action_space_cls="discrete")
    action_dist_params = None
    num_actions, max_action, action_dist_params = get_action_params(
        env.action_space,
        action_dist_params,
    )
    assert action_dist_params["cls"] == "Categorical"
    action_dist_cls = __Distribution__.build(action_dist_params)

    T, B = 20, 4
    learner_outputs_policy_logits = torch.randn(T, B, num_actions)

    if test_entropy:
        target_policy_action_dist = action_dist_cls(logits=learner_outputs_policy_logits)

        entropy_loss = -target_policy_action_dist.entropy().mean()
        _entropy_loss = _compute_entropy_loss(learner_outputs_policy_logits)
        assert torch.all(torch.isclose(entropy_loss, _entropy_loss))

        # --- also test flattened (T * B, -1)
        target_policy_action_dist = action_dist_cls(logits=learner_outputs_policy_logits.view(T * B, -1))
        entropy_loss = -target_policy_action_dist.entropy().mean()
        _entropy_loss = _compute_entropy_loss(learner_outputs_policy_logits.view(T * B, -1))
        assert torch.all(torch.isclose(entropy_loss, _entropy_loss))

    actions = action_dist_cls(logits=learner_outputs_policy_logits).sample()  # use sample instead of rsample for Categorical

    if test_pg:
        target_policy_action_dist = action_dist_cls(logits=learner_outputs_policy_logits)
        value_error = torch.randn((T, B))
        target_action_log_probs = target_policy_action_dist.log_prob(actions)

        pg_loss = -torch.mean(target_action_log_probs * value_error.detach())  # policy gradient
        _pg_loss = _compute_policy_gradient_loss(learner_outputs_policy_logits, actions, value_error)
        assert torch.all(torch.isclose(pg_loss, _pg_loss))

    if test_kl:
        target_policy_action_dist = action_dist_cls(logits=learner_outputs_policy_logits)

        actor_outputs_policy_logits = torch.randn(T, B, num_actions)
        behavior_policy_action_dist = action_dist_cls(logits=actor_outputs_policy_logits)

        behavior_action_log_probs = behavior_policy_action_dist.log_prob(actions)
        target_action_log_probs = target_policy_action_dist.log_prob(actions)
        kl = behavior_action_log_probs - target_action_log_probs
        kl_loss = torch.mean(kl)  # noqa


def test_losses_tanhnormal(test_entropy=True):
    env = MockEnv(obs_space_type="1d", action_space_cls="box")
    action_dist_params = None
    num_actions, max_action, action_dist_params = get_action_params(
        env.action_space,
        action_dist_params,
    )
    assert action_dist_params["cls"] == "SquashedDiagGaussian"
    action_dist_cls = __Distribution__.build(action_dist_params)

    if test_entropy:
        T, B = 20, 4
        learner_outputs_policy_logits = (torch.randn(T, B, num_actions), torch.randn(T, B, num_actions).exp())
        loc, scale = learner_outputs_policy_logits

        target_policy_action_dist = action_dist_cls(loc, scale)
        entropy_loss = -target_policy_action_dist.entropy().mean()

        # --- also test flattened (T * B, -1)
        target_policy_action_dist1 = action_dist_cls(loc.view(T * B, -1), scale.view(T * B, -1))
        entropy_loss1 = -target_policy_action_dist1.entropy().mean()
        assert torch.all(torch.isclose(entropy_loss, entropy_loss1))



# Taken from https://github.com/facebookresearch/moolib/blob/e8b2de7ac5df3a9b3ee2548a33f61100a95152ef/examples/common/vtrace.py#L156

# This file taken from
#     https://github.com/deepmind/scalable_agent/blob/
#         cd66d00914d56c8ba2f0615d9cdeefcb169a8d70/vtrace.py
# and modified.

# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions to compute V-trace off-policy actor critic targets.

For details and theory see:

"IMPALA: Scalable Distributed Deep-RL with
Importance Weighted Actor-Learner Architectures"
by Espeholt, Soyer, Munos et al.

See https://arxiv.org/abs/1802.01561 for the full paper.
"""

import collections

import torch

VTraceReturns = collections.namedtuple("VTraceReturns", ["vs", "pg_advantages"])


@torch.no_grad()
def from_importance_weights(
    log_rhos,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
    lambda_=1.0,
):
    r"""V-trace from log importance weights.

    Calculates V-trace actor critic targets as described in

    "IMPALA: Scalable Distributed Deep-RL with
    Importance Weighted Actor-Learner Architectures"
    by Espeholt, Soyer, Munos et al.

    In the notation used throughout documentation and comments, T refers to the
    time dimension ranging from 0 to T-1. B refers to the batch size and
    NUM_ACTIONS refers to the number of actions. This code also supports the
    case where all tensors have the same number of additional dimensions, e.g.,
    `rewards` is [T, B, C], `values` is [T, B, C], `bootstrap_value` is [B, C].

    Args:
      log_rhos: A float32 tensor of shape [T, B, NUM_ACTIONS] representing the log
        importance sampling weights, i.e.
        log(target_policy(a) / behaviour_policy(a)). V-trace performs operations
        on rhos in log-space for numerical stability.
      discounts: A float32 tensor of shape [T, B] with discounts encountered when
        following the behaviour policy.
      rewards: A float32 tensor of shape [T, B] containing rewards generated by
        following the behaviour policy.
      values: A float32 tensor of shape [T, B] with the value function estimates
        wrt. the target policy.
      bootstrap_value: A float32 of shape [B] with the value function estimate at
        time T.
      clip_rho_threshold: A scalar float32 tensor with the clipping threshold for
        importance weights (rho) when calculating the baseline targets (vs).
        rho^bar in the paper. If None, no clipping is applied.
      clip_pg_rho_threshold: A scalar float32 tensor with the clipping threshold
        on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)). If
        None, no clipping is applied.
      lambda_: Mix between 1-step (lambda_=0) and n-step (lambda_=1). See Remark 2
        in paper. Defaults to lambda_=1.
      name: The name scope that all V-trace operations will be created in.

    Returns:
      A VTraceReturns namedtuple (vs, pg_advantages) where:
        vs: A float32 tensor of shape [T, B]. Can be used as target to
          train a baseline (V(x_t) - vs_t)^2.
        pg_advantages: A float32 tensor of shape [T, B]. Can be used as the
          advantage in the calculation of policy gradients.
    """
    # from https://github.com/ray-project/ray/blob/c61910487fc01efefed6ef759d461970c7b2f974/rllib/agents/impala/vtrace_torch.py#L308
    # Make sure tensor ranks are consistent.
    rho_rank = len(log_rhos.size())  # Usually 2.
    assert rho_rank == len(values.size())
    assert rho_rank - 1 == len(bootstrap_value.size()), "must have rank {}".format(rho_rank - 1)
    assert rho_rank == len(discounts.size())
    assert rho_rank == len(rewards.size())

    rhos = torch.exp(log_rhos)
    if clip_rho_threshold is not None:
        clipped_rhos = torch.clamp(rhos, max=clip_rho_threshold)
    else:
        clipped_rhos = rhos

    cs = torch.clamp(rhos, max=1.0)
    if lambda_ is not None:
        cs *= lambda_
    # Append bootstrapped value to get [v1, ..., v_t+1]
    values_t_plus_1 = torch.cat([values[1:], torch.unsqueeze(bootstrap_value, 0)], dim=0)
    deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

    acc = torch.zeros_like(bootstrap_value)
    result = []
    for t in range(discounts.shape[0] - 1, -1, -1):
        acc = deltas[t] + discounts[t] * cs[t] * acc
        result.append(acc)
    result.reverse()
    vs_minus_v_xs = torch.stack(result)

    # Add V(x_s) to get v_s.
    vs = torch.add(vs_minus_v_xs, values)

    # Advantage for policy gradient.
    broadcasted_bootstrap_values = torch.ones_like(vs[0]) * bootstrap_value
    vs_t_plus_1 = torch.cat([vs[1:], broadcasted_bootstrap_values.unsqueeze(0)], dim=0)
    if clip_pg_rho_threshold is not None:
        clipped_pg_rhos = torch.clamp(rhos, max=clip_pg_rho_threshold)
    else:
        clipped_pg_rhos = rhos
    pg_advantages = clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values)

    # Make sure no gradients backpropagated through the returned values.
    return VTraceReturns(vs=vs, pg_advantages=pg_advantages)



# modified from https://github.com/facebookresearch/moolib/blob/e8b2de7ac5df3a9b3ee2548a33f61100a95152ef/examples/vtrace/experiment.py
# and https://github.com/facebookresearch/minihack/blob/c17084885833cbeee8bdd6684d0cde0a2536e3bb/minihack/agent/polybeast/polyhydra.py

# Copyright (c) Facebook, Inc. and its affiliates.
import copy
import getpass
import logging
import os
import pprint
import signal
import socket
import time

import coolname
import hydra
import moolib
import omegaconf
import torch
from moolib.examples import common
from moolib.examples.common import nest, record
from tinyspace import sample_from_space

from multibeast.builder import __Agent__, __MakeEnv__
from multibeast.envpool import EnvBatchState, EnvPool

from .record_utils import calculate_sps, load_checkpoint, log, save_checkpoint

SLUG = None
# need to cache this, otherwise uid will differ b/w what shows up in logger and hydra config
# when calling multihydra.run() from a different file with @hydra.main() decorator.
def uid():
    global SLUG
    if SLUG is None:
        SLUG = coolname.generate_slug(2)
    return "%s:%i:%s" % (socket.gethostname(), os.getpid(), SLUG)


omegaconf.OmegaConf.register_new_resolver("uid", uid, use_cache=True)


def run(cfg: omegaconf.DictConfig):  # noqa: C901
    global FLAGS
    FLAGS = cfg

    if not os.path.isabs(FLAGS.savedir):
        FLAGS.savedir = os.path.join(hydra.utils.get_original_cwd(), FLAGS.savedir)

    logging.info("flags:\n%s\n", pprint.pformat(dict(FLAGS)))

    if record.symlink_path(FLAGS.savedir, os.path.join(hydra.utils.get_original_cwd(), "latest")):
        logging.info("savedir: %s (symlinked as 'latest')", FLAGS.savedir)
    else:
        logging.info("savedir: %s", FLAGS.savedir)

    train_id = "%s/%s/%s" % (
        FLAGS.entity if FLAGS.entity is not None else getpass.getuser(),
        FLAGS.project,
        FLAGS.group,
    )

    logging.info("train_id: %s", train_id)

    if FLAGS.use_moolib_envpool:
        EnvPoolCls = moolib.EnvPool  # only supports discrete action space curently
    else:
        EnvPoolCls = EnvPool

    create_env_fn = __MakeEnv__.build(FLAGS.make_env_name, FLAGS.env)

    envs = EnvPoolCls(
        create_env_fn,
        num_processes=FLAGS.num_actor_processes,
        batch_size=FLAGS.actor_batch_size,
        num_batches=FLAGS.num_actor_batches,
    )

    logging.info(f"EnvPool started: {envs}")

    dummy_env = create_env_fn()  # TODO: pass an option `dummy_env=True`, so that only the desired attributes can be accessed
    observation_space = dummy_env.observation_space
    action_space = dummy_env.action_space
    info_keys_custom = getattr(dummy_env, "info_keys_custom", [])
    dummy_env.close()
    del dummy_env
    logging.info(f"observation_space: {observation_space}")
    logging.info(f"action_space: {action_space}")

    Agent = __Agent__.get(FLAGS.get("agent", "Impala"))
    model, learner_state = Agent.create_agent(FLAGS, observation_space, action_space)
    model.to(device=FLAGS.device)

    model_numel = sum(p.numel() for p in model.parameters() if p.requires_grad)
    logging.info("Number of model parameters: %i", model_numel)
    record.write_metadata(
        FLAGS.localdir,
        hydra.utils.get_original_cwd(),
        flags=omegaconf.OmegaConf.to_container(FLAGS),
        model_numel=model_numel,
    )

    if FLAGS.wandb:
        import wandb

        wandb.init(
            project=str(FLAGS.project),
            config=omegaconf.OmegaConf.to_container(FLAGS),
            group=FLAGS.group,
            entity=FLAGS.entity,
            name=FLAGS.local_name,
            tags=FLAGS.get("tags", None),
            notes=FLAGS.get("notes", None),
        )

    zero_action = sample_from_space(action_space, batch_size=FLAGS.actor_batch_size, to_torch_tensor=True)
    env_states = [
        EnvBatchState(
            FLAGS,
            model,
            zero_action,
            info_keys_custom=info_keys_custom,
        )
        for _ in range(FLAGS.num_actor_batches)
    ]

    rpc = moolib.Rpc()
    rpc.set_name(FLAGS.local_name)
    rpc.connect(FLAGS.connect)

    rpc_group = moolib.Group(rpc, name=train_id)

    accumulator = moolib.Accumulator(
        group=rpc_group,
        name="model",
        parameters=model.parameters(),
        buffers=model.buffers(),
    )
    accumulator.set_virtual_batch_size(FLAGS.virtual_batch_size)

    learn_batcher = moolib.Batcher(FLAGS.batch_size, FLAGS.device, dim=1)

    agent_stats = Agent.create_stats(FLAGS)
    runner_stats = {
        "SPS": common.StatMean(),
        "env_act_steps": common.StatSum(),
        "env_train_steps": common.StatSum(),
        "steps_done": common.StatSum(),
        "episodes_done": common.StatSum(),
        #
        "mean_episode_return": common.StatMean(),
        "mean_episode_step": common.StatMean(),
        "running_reward": common.StatMean(),
        "running_step": common.StatMean(),
        "end_episode_success": common.StatMean(),
        "end_episode_progress": common.StatMean(),
        #
        "virtual_batch_size": common.StatMean(),
        "num_gradients": common.StatMean(),
        #
        "optimizer_steps": common.StatSum(),  # should be updated by `Agent.step_optimizer()`
        "model_version": common.StatSum(),  # should be updated by `Agent.step_optimizer()`
    }
    for k in info_keys_custom:
        runner_stats[f"end_{k}"] = common.StatMean()

    stats = dict(**runner_stats, **agent_stats)
    learner_state.global_stats = copy.deepcopy(stats)

    checkpoint_path = os.path.join(FLAGS.savedir, "checkpoint.tar")
    if os.path.exists(checkpoint_path):
        logging.info("Loading checkpoint: %s" % checkpoint_path)
        load_checkpoint(checkpoint_path, learner_state)
        accumulator.set_model_version(learner_state.model_version)
        logging.info("loaded stats %s", learner_state.global_stats)

    global_stats_accumulator = common.GlobalStatsAccumulator(rpc_group, learner_state.global_stats)

    terminate = False
    previous_signal_handler = {}

    def signal_handler(signum, frame):
        nonlocal terminate
        logging.info(
            "Got signal %s, quitting!",
            signal.strsignal(signum) if hasattr(signal, "strsignal") else signum,
        )
        terminate = True
        previous_handler = previous_signal_handler[signum]
        if previous_handler is not None:
            previous_signal_handler[signum] = None
            signal.signal(signum, previous_handler)

    previous_signal_handler[signal.SIGTERM] = signal.signal(signal.SIGTERM, signal_handler)
    previous_signal_handler[signal.SIGINT] = signal.signal(signal.SIGINT, signal_handler)

    if torch.backends.cudnn.is_available():
        logging.info("Optimising CuDNN kernels")
        torch.backends.cudnn.benchmark = True

    # Run.
    now = time.time()
    warm_up_time = FLAGS.warmup
    prev_env_train_steps = 0
    prev_global_env_train_steps = 0
    next_env_index = 0
    last_log = now
    last_reduce_stats = now
    is_leader = False
    is_connected = False
    while not terminate:
        prev_now = now
        now = time.time()

        steps = learner_state.global_stats["env_train_steps"].result()
        if steps >= FLAGS.total_steps:
            logging.info("Stopping training after %i steps", steps)
            break

        rpc_group.update()
        accumulator.update()
        if accumulator.wants_state():
            assert accumulator.is_leader()
            accumulator.set_state(learner_state.save())
        if accumulator.has_new_state():
            assert not accumulator.is_leader()
            learner_state.load(accumulator.state())

        was_connected = is_connected
        is_connected = accumulator.connected()
        if not is_connected:
            if was_connected:
                logging.warning("Training interrupted!")
            # If we're not connected, sleep for a bit so we don't busy-wait
            logging.info("Your training will commence shortly.")
            time.sleep(1)
            continue

        was_leader = is_leader
        is_leader = accumulator.is_leader()
        if not was_connected:
            logging.info(
                "Training started. Leader is %s, %d members, model version is %d"
                % (
                    "me!" if is_leader else accumulator.get_leader(),
                    len(rpc_group.members()),
                    learner_state.model_version,
                )
            )
            prev_global_env_train_steps = learner_state.global_stats["env_train_steps"].result()

            if warm_up_time > 0:
                logging.info("Warming up for %g seconds", warm_up_time)

        if warm_up_time > 0:
            warm_up_time -= now - prev_now

        learner_state.train_time += now - prev_now
        if now - last_reduce_stats >= 2:
            last_reduce_stats = now
            # NOTE: If getting "TypeError: unsupported operand type(s) for -: 'float' and 'StatMean'"
            # then probably assigning with `stats["key"] = value`. Use `stats["key"] += value` instead.
            global_stats_accumulator.reduce(stats)
        if now - last_log >= FLAGS.log_interval:
            delta = now - last_log
            last_log = now

            global_stats_accumulator.reduce(stats)
            global_stats_accumulator.reset()

            prev_env_train_steps = calculate_sps(stats, delta, prev_env_train_steps, is_global=False)
            prev_global_env_train_steps = calculate_sps(
                learner_state.global_stats, delta, prev_global_env_train_steps, is_global=True
            )

            steps = learner_state.global_stats["env_train_steps"].result()

            log(stats, step=steps, is_global=False, wandb=FLAGS.wandb)
            log(learner_state.global_stats, step=steps, is_global=True, wandb=FLAGS.wandb)

            if warm_up_time > 0:
                logging.info("Warming up up for an additional %g seconds", round(warm_up_time))

        if is_leader:
            if not was_leader:
                leader_filename = os.path.join(FLAGS.savedir, "leader-%03d" % learner_state.num_previous_leaders)
                record.symlink_path(FLAGS.localdir, leader_filename)
                logging.info("Created symlink %s -> %s", leader_filename, FLAGS.localdir)
                learner_state.num_previous_leaders += 1
            if not was_leader and not os.path.exists(checkpoint_path):
                logging.info("Training a new model from scratch.")
            if learner_state.train_time - learner_state.last_checkpoint >= FLAGS.checkpoint_interval:
                learner_state.last_checkpoint = learner_state.train_time
                save_checkpoint(checkpoint_path, learner_state, FLAGS)
            if learner_state.train_time - learner_state.last_checkpoint_history >= FLAGS.checkpoint_history_interval:
                learner_state.last_checkpoint_history = learner_state.train_time
                save_checkpoint(
                    os.path.join(
                        FLAGS.savedir,
                        "checkpoint_v%d.tar" % learner_state.model_version,
                        FLAGS,
                    ),
                    learner_state,
                )

        if accumulator.has_gradients():
            gradient_stats = accumulator.get_gradient_stats()
            stats["virtual_batch_size"] += gradient_stats["batch_size"]
            stats["num_gradients"] += gradient_stats["num_gradients"]
            Agent.step_optimizer(FLAGS, learner_state, stats)
            accumulator.zero_gradients()
        elif not learn_batcher.empty() and accumulator.wants_gradients():
            Agent.compute_gradients(FLAGS, learn_batcher.get(), learner_state, stats)
            stats["env_train_steps"] += FLAGS.unroll_length * FLAGS.batch_size
            accumulator.reduce_gradients(FLAGS.batch_size)
        else:
            if accumulator.wants_gradients():
                accumulator.skip_gradients()

            # Generate data.
            cur_index = next_env_index
            next_env_index = (next_env_index + 1) % FLAGS.num_actor_batches

            env_state = env_states[cur_index]
            if env_state.future is None:  # need to initialize
                env_state.future = envs.step(cur_index, env_state.prev_action)
            cpu_env_outputs = env_state.future.result()

            env_outputs = nest.map(lambda t: t.to(FLAGS.device, copy=True), cpu_env_outputs)

            env_outputs["prev_action"] = env_state.prev_action
            prev_core_state = env_state.core_state
            model.eval()
            with torch.no_grad():
                actor_outputs, env_state.core_state = model(
                    nest.map(lambda t: t.unsqueeze(0), env_outputs),
                    env_state.core_state,
                )
            actor_outputs = nest.map(lambda t: t.squeeze(0), actor_outputs)
            action = actor_outputs["action"]
            env_state.update(cpu_env_outputs, action, stats)
            # envs.step invalidates cpu_env_outputs
            del cpu_env_outputs
            env_state.future = envs.step(cur_index, action)

            stats["env_act_steps"] += action.numel()

            last_data = {
                "env_outputs": env_outputs,
                "actor_outputs": actor_outputs,
            }
            if warm_up_time <= 0:
                env_state.time_batcher.stack(last_data)

            if not env_state.time_batcher.empty():
                data = env_state.time_batcher.get()
                data["initial_core_state"] = env_state.initial_core_state
                learn_batcher.cat(data)

                # We need the last entry of the previous time batch
                # to be put into the first entry of this time batch,
                # with the initial_core_state to match
                env_state.initial_core_state = prev_core_state
                env_state.time_batcher.stack(last_data)
    if is_connected and is_leader:
        save_checkpoint(checkpoint_path, learner_state, FLAGS)
    logging.info("Graceful exit. Bye bye!")


# Override config_path via --config_path.
@hydra.main(config_path=".", config_name="config")
def main(cfg: omegaconf.DictConfig):
    run(cfg)


if __name__ == "__main__":
    # moolib.set_log_level("debug")
    # moolib.set_max_threads(1)
    main()

# see https://github.com/facebookresearch/moolib/tree/main/examples#fully-fledged-vtrace-agent
# for usage



import numpy as np


class MockEnv:
    def __init__(self, obs_space_type="rgb", action_space_cls="discrete"):
        if obs_space_type == "rgb":
            self._observation_space = dict(
                shape=(3, 64, 64),
                low=0,
                high=255,
                dtype=np.uint8,
            )
        elif obs_space_type == "1d":
            self._observation_space = dict(
                shape=(100,),
                low=-np.inf,
                high=np.inf,
                dtype=np.float32,
            )
        else:
            raise ValueError

        if action_space_cls == "discrete":
            action_space = dict(
                cls="discrete",
                shape=(),
                low=0,
                high=10,
                dtype=np.int64,
            )
        elif action_space_cls == "box":
            action_space = dict(
                cls="box",
                shape=(10,),
                low=-1.0,
                high=1.0,
                dtype=np.int64,
            )
        else:
            raise ValueError

        self._action_space = action_space

    @property
    def info_keys_custom(self):
        return ["progress", "success", "custom_metric"]

    @property
    def max_episode_steps(self):
        return 20

    @property
    def observation_space(self):
        return self._observation_space

    @property
    def action_space(self):
        return self._action_space

    def reset(self):
        pass

    def step(self, action):
        pass



# This file taken from
#     https://github.com/deepmind/scalable_agent/blob/
#         cd66d00914d56c8ba2f0615d9cdeefcb169a8d70/vtrace.py
# and modified.

# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions to compute V-trace off-policy actor critic targets.

For details and theory see:

"IMPALA: Scalable Distributed Deep-RL with
Importance Weighted Actor-Learner Architectures"
by Espeholt, Soyer, Munos et al.

See https://arxiv.org/abs/1802.01561 for the full paper.
"""

import collections

import torch
import torch.nn.functional as F


VTraceFromLogitsReturns = collections.namedtuple(
    "VTraceFromLogitsReturns",
    [
        "vs",
        "pg_advantages",
        "log_rhos",
        "behavior_action_log_probs",
        "target_action_log_probs",
    ],
)

VTraceReturns = collections.namedtuple("VTraceReturns", "vs pg_advantages")


def action_log_probs(policy_logits, actions):
    """Computes action log-probs from policy logits and actions.

    In the notation used throughout documentation and comments, T refers to the
    time dimension ranging from 0 to T-1. B refers to the batch size and
    NUM_ACTIONS refers to the number of actions.

    Args:
      policy_logits: A float32 tensor of shape [T, B, NUM_ACTIONS] with
        un-normalized log-probabilities parameterizing a softmax policy.
      actions: An int32 tensor of shape [T, B] with actions.

    Returns:
      A float32 tensor of shape [T, B] corresponding to the sampling log
      probability of the chosen action w.r.t. the policy.
    """
    return -F.nll_loss(
        F.log_softmax(torch.flatten(policy_logits, 0, -2), dim=-1),
        torch.flatten(actions),
        reduction="none",
    ).view_as(actions)


def from_logits(
    behavior_policy_logits,
    target_policy_logits,
    actions,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    r"""V-trace for softmax policies.

    Calculates V-trace actor critic targets for softmax polices as described in

    "IMPALA: Scalable Distributed Deep-RL with
    Importance Weighted Actor-Learner Architectures"
    by Espeholt, Soyer, Munos et al.

    Target policy refers to the policy we are interested in improving and
    behaviour policy refers to the policy that generated the given
    rewards and actions.

    In the notation used throughout documentation and comments, T refers to the
    time dimension ranging from 0 to T-1. B refers to the batch size and
    NUM_ACTIONS refers to the number of actions.

    Args:
      behaviour_policy_logits: A float32 tensor of shape [T, B, NUM_ACTIONS] with
        un-normalized log-probabilities parametrizing the softmax behaviour
        policy.
      target_policy_logits: A float32 tensor of shape [T, B, NUM_ACTIONS] with
        un-normalized log-probabilities parametrizing the softmax target policy.
      actions: An int32 tensor of shape [T, B] of actions sampled from the
        behaviour policy.
      discounts: A float32 tensor of shape [T, B] with the discount encountered
        when following the behaviour policy.
      rewards: A float32 tensor of shape [T, B] with the rewards generated by
        following the behaviour policy.
      values: A float32 tensor of shape [T, B] with the value function estimates
        wrt. the target policy.
      bootstrap_value: A float32 of shape [B] with the value function estimate at
        time T.
      clip_rho_threshold: A scalar float32 tensor with the clipping threshold for
        importance weights (rho) when calculating the baseline targets (vs).
        rho^bar in the paper.
      clip_pg_rho_threshold: A scalar float32 tensor with the clipping threshold
        on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)).
      name: The name scope that all V-trace operations will be created in.

    Returns:
      A `VTraceFromLogitsReturns` namedtuple with the following fields:
        vs: A float32 tensor of shape [T, B]. Can be used as target to train a
            baseline (V(x_t) - vs_t)^2.
        pg_advantages: A float 32 tensor of shape [T, B]. Can be used as an
          estimate of the advantage in the calculation of policy gradients.
        log_rhos: A float32 tensor of shape [T, B] containing the log importance
          sampling weights (log rhos).
        behaviour_action_log_probs: A float32 tensor of shape [T, B] containing
          behaviour policy action log probabilities (log \mu(a_t)).
        target_action_log_probs: A float32 tensor of shape [T, B] containing
          target policy action probabilities (log \pi(a_t)).
    """
    target_action_log_probs = action_log_probs(target_policy_logits, actions)
    behavior_action_log_probs = action_log_probs(behavior_policy_logits, actions)
    log_rhos = target_action_log_probs - behavior_action_log_probs
    vtrace_returns = from_importance_weights(
        log_rhos=log_rhos,
        discounts=discounts,
        rewards=rewards,
        values=values,
        bootstrap_value=bootstrap_value,
        clip_rho_threshold=clip_rho_threshold,
        clip_pg_rho_threshold=clip_pg_rho_threshold,
    )
    return VTraceFromLogitsReturns(
        log_rhos=log_rhos,
        behavior_action_log_probs=behavior_action_log_probs,
        target_action_log_probs=target_action_log_probs,
        **vtrace_returns._asdict(),
    )


@torch.no_grad()
def from_importance_weights(
    log_rhos,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
):
    r"""V-trace from log importance weights.

    Calculates V-trace actor critic targets as described in

    "IMPALA: Scalable Distributed Deep-RL with
    Importance Weighted Actor-Learner Architectures"
    by Espeholt, Soyer, Munos et al.

    In the notation used throughout documentation and comments, T refers to the
    time dimension ranging from 0 to T-1. B refers to the batch size and
    NUM_ACTIONS refers to the number of actions. This code also supports the
    case where all tensors have the same number of additional dimensions, e.g.,
    `rewards` is [T, B, C], `values` is [T, B, C], `bootstrap_value` is [B, C].

    Args:
      log_rhos: A float32 tensor of shape [T, B, NUM_ACTIONS] representing the log
        importance sampling weights, i.e.
        log(target_policy(a) / behaviour_policy(a)). V-trace performs operations
        on rhos in log-space for numerical stability.
      discounts: A float32 tensor of shape [T, B] with discounts encountered when
        following the behaviour policy.
      rewards: A float32 tensor of shape [T, B] containing rewards generated by
        following the behaviour policy.
      values: A float32 tensor of shape [T, B] with the value function estimates
        wrt. the target policy.
      bootstrap_value: A float32 of shape [B] with the value function estimate at
        time T.
      clip_rho_threshold: A scalar float32 tensor with the clipping threshold for
        importance weights (rho) when calculating the baseline targets (vs).
        rho^bar in the paper. If None, no clipping is applied.
      clip_pg_rho_threshold: A scalar float32 tensor with the clipping threshold
        on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)). If
        None, no clipping is applied.
      name: The name scope that all V-trace operations will be created in.

    Returns:
      A VTraceReturns namedtuple (vs, pg_advantages) where:
        vs: A float32 tensor of shape [T, B]. Can be used as target to
          train a baseline (V(x_t) - vs_t)^2.
        pg_advantages: A float32 tensor of shape [T, B]. Can be used as the
          advantage in the calculation of policy gradients.
    """
    rhos = torch.exp(log_rhos)
    if clip_rho_threshold is not None:
        clipped_rhos = torch.clamp(rhos, max=clip_rho_threshold)
    else:
        clipped_rhos = rhos

    cs = torch.clamp(rhos, max=1.0)
    # Append bootstrapped value to get [v1, ..., v_t+1]
    values_t_plus_1 = torch.cat(
        [values[1:], torch.unsqueeze(bootstrap_value, 0)], dim=0
    )
    deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

    acc = torch.zeros_like(bootstrap_value)
    result = []
    for t in range(discounts.shape[0] - 1, -1, -1):
        acc = deltas[t] + discounts[t] * cs[t] * acc
        result.append(acc)
    result.reverse()
    vs_minus_v_xs = torch.stack(result)

    # Add V(x_s) to get v_s.
    vs = torch.add(vs_minus_v_xs, values)

    # Advantage for policy gradient.
    broadcasted_bootstrap_values = torch.ones_like(vs[0]) * bootstrap_value
    vs_t_plus_1 = torch.cat([vs[1:], broadcasted_bootstrap_values.unsqueeze(0)], dim=0)
    if clip_pg_rho_threshold is not None:
        clipped_pg_rhos = torch.clamp(rhos, max=clip_pg_rho_threshold)
    else:
        clipped_pg_rhos = rhos
    pg_advantages = clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values)

    # Make sure no gradients backpropagated through the returned values.
    return VTraceReturns(vs=vs, pg_advantages=pg_advantages)



from functools import partial
from typing import Callable

import omegaconf

from .utils.registry import Registry


def parse_params(params):
    assert "cls" in params.keys()
    if isinstance(params, omegaconf.DictConfig):
        # deal with "omegaconf.errors.ConfigTypeError: DictConfig in struct mode does not support pop"
        params = dict(params)
    else:
        params = params.copy()
    c = params.pop("cls")
    return c, params


def build_make_env(make_env_name, make_env_params) -> Callable:
    r"""This function gets a function in the registry, which can be called to make an environment.

    Example:
        ```python
        from multibeast.builder import __MakeEnv__

        @MakeEnvRegistry.register()
        def make_env(**kwargs):  # kwargs come from hydra_cfg.yaml
            env = gym.make("CartPole", **kwargs)
            return env

        # where `make_env_name = "make_env" == make_env.__name__`
        ```
    """
    create_env_fn = lambda: __MakeEnv__.get(make_env_name)(make_env_params)
    return create_env_fn


def build_feature_extractor(feature_extractor_params, observation_space, action_space):
    c, params = parse_params(feature_extractor_params)
    FeatureExtractorCls = __FeatureExtractor__.get(c)
    return FeatureExtractorCls(observation_space, action_space, **params)


def build_policy_net(policy_params, *args):
    if policy_params is None:
        policy_params = dict(cls="PolicyNet")

    c, params = parse_params(policy_params)
    PolicyNetCls = __PolicyNet__.get(c)
    return PolicyNetCls(*args, **params)


def build_distribution(distribution_params):
    r"""Returns a wrapped constructor for a torch.Distribution."""
    c, params = parse_params(distribution_params)
    try:
        # import module to load register() calls
        import multibeast.distributions  # noqa: F401

        DistributionCls = __Distribution__.get(c)
    except KeyError:
        import torch.distributions as D

        DistributionCls = getattr(D, c)
    return partial(DistributionCls, **params)


def build_optimizer(optimizer_params, model_params):
    c, params = parse_params(optimizer_params)

    import torch.optim  # noqa: F401

    OptimizerCls = getattr(torch.optim, c)
    return OptimizerCls(model_params, **params)


def build_agent(agent_name):
    AgentCls = __Agent__.get(agent_name)
    return AgentCls


__MakeEnv__ = Registry("MakeEnv")
__MakeEnv__.build = build_make_env

__FeatureExtractor__ = Registry("FeatureExtractor")
__FeatureExtractor__.build = build_feature_extractor

__PolicyNet__ = Registry("PolicyNet")
__PolicyNet__.build = build_policy_net

__Distribution__ = Registry("Distribution")
__Distribution__.build = build_distribution

__Optimizer__ = Registry("Optimizer")
__Optimizer__.build = build_optimizer

__Agent__ = Registry("Agent")
__Agent__.build = build_agent



# modified from https://github.com/facebookresearch/moolib/blob/e8b2de7ac5df3a9b3ee2548a33f61100a95152ef/examples/common/__init__.py#L154

import moolib
import torch
from moolib.examples.common import RunningMeanStd


class EnvBatchState:
    r"""Holds state for a batch of environments.

    Also tracks ["success", "progress"] at the end of the episode, if those keys are in the info dict.

    Args:
        FLAGS:
          unroll_length: Length of a rollout (i.e., number of steps that an actor has
            to be perform before sending its experience to the learner).
        model:
        zero_action: a tensor of zeros with shape [`FLAGS.actor_batch_size`, `env.action_space.action_dim`]
        info_custom_keys: a list of keys to track stats for from the info dict.
    """

    def __init__(self, FLAGS, model, zero_action, with_scaled_reward: bool = False, info_keys_custom: list = None):
        batch_size = FLAGS.actor_batch_size
        device = FLAGS.device
        self.batch_size = batch_size
        self.prev_action = zero_action.clone().to(device)
        self.future = None
        self.core_state = model.initial_state(batch_size=batch_size)
        self.core_state = tuple(s.to(device) for s in self.core_state)
        self.initial_core_state = self.core_state

        self.running_reward = torch.zeros(batch_size)
        self.step_count = torch.zeros(batch_size)

        self.with_scaled_reward = with_scaled_reward
        if with_scaled_reward:
            self.discounting = FLAGS.model.discounting
            self.weighted_returns = torch.zeros(batch_size)
            self.weighted_returns_rms = RunningMeanStd()

        if info_keys_custom is None:
            info_keys_custom = []
        self.info_keys_custom = info_keys_custom

        self.time_batcher = moolib.Batcher(FLAGS.unroll_length + 1, device)

    def update(self, env_outputs, action, stats):
        self.prev_action = action
        self.running_reward += env_outputs["reward"]

        if self.with_scaled_reward:
            self.weighted_returns *= self.discounting
            self.weighted_returns += env_outputs["reward"]
            self.weighted_returns_rms.update(self.weighted_returns)

            self.scaled_reward = env_outputs["reward"] / torch.sqrt(self.weighted_returns_rms.var + 1e-8)

        self.step_count += 1

        info = env_outputs.get("info", {})

        # TODO: https://arxiv.org/abs/1712.00378
        # timeout = info["timeout"]
        done = env_outputs["done"]

        episode_return = self.running_reward * done
        episode_step = self.step_count * done

        episodes_done = done.sum().item()
        if episodes_done > 0:
            stats["mean_episode_return"] += episode_return.sum().item() / episodes_done
            stats["mean_episode_step"] += episode_step.sum().item() / episodes_done
            if "success" in info.keys():
                stats["end_episode_success"] += (info["success"] * done).sum().item() / episodes_done
            if "progress" in info.keys():
                stats["end_episode_progress"] += (info["progress"] * done).sum().item() / episodes_done

            for k in self.info_keys_custom:
                stats[f"end_{k}"] += (info[k] * done).sum().item() / episodes_done

        stats["steps_done"] += done.numel()
        stats["episodes_done"] += episodes_done

        stats["running_reward"] += self.running_reward.mean().item()
        stats["running_step"] += self.step_count.mean().item()

        not_done = ~done

        self.running_reward *= not_done
        self.step_count *= not_done
        if self.with_scaled_reward:
            self.weighted_returns *= not_done



import dataclasses
import logging
from typing import Optional

import torch
import torch.nn as nn
from moolib.examples import common
from moolib.examples.common import nest

from multibeast.builder import __Agent__, __FeatureExtractor__, __Optimizer__

from .impala_net import ImpalaNet
from .vtrace import from_importance_weights


@dataclasses.dataclass
class ImpalaLearnerState:
    model: torch.nn.Module
    optimizer: torch.optim.Optimizer
    scheduler: torch.optim.lr_scheduler.LambdaLR
    model_version: int = 0
    num_previous_leaders: int = 0
    train_time: float = 0
    last_checkpoint: float = 0
    last_checkpoint_history: float = 0
    global_stats: Optional[dict] = None

    def save(self):
        r = dataclasses.asdict(self)
        r["model"] = self.model.state_dict()
        r["optimizer"] = self.optimizer.state_dict()
        r["scheduler"] = self.scheduler.state_dict()
        return r

    def load(self, state):
        for k, v in state.items():
            if k not in ("model", "optimizer", "scheduler", "global_stats"):
                setattr(self, k, v)
        self.model_version = state["model_version"]
        self.model.load_state_dict(state["model"])
        self.optimizer.load_state_dict(state["optimizer"])
        self.scheduler.load_state_dict(state["scheduler"])

        for k, v in state["global_stats"].items():
            if k in self.global_stats:
                self.global_stats[k] = type(self.global_stats[k])(**v)


@__Agent__.register()
class Impala:
    def __new__(cls):
        raise TypeError("Static class cannot be instantiated")

    @staticmethod
    def _create_model(FLAGS, observation_space, action_space):
        if FLAGS.feature_extractor:
            feature_extractor = __FeatureExtractor__.build(
                FLAGS.feature_extractor,
                observation_space,
                action_space,
            )
        else:
            feature_extractor = None

        MFLAGS = FLAGS.model
        action_dist_params = MFLAGS.get("action_dist_params", None)
        policy_params = MFLAGS.get("policy_params", dict(cls="PolicyNet"))

        model = ImpalaNet(
            observation_space,
            action_space,
            feature_extractor=feature_extractor,
            action_dist_params=action_dist_params,
            policy_params=policy_params,
            use_lstm=MFLAGS.use_lstm,
        )

        logging.info(f"model: \n{model}")

        return model

    @staticmethod
    def _create_scheduler(FLAGS, optimizer):
        factor = FLAGS.unroll_length * FLAGS.virtual_batch_size / FLAGS.total_steps
        scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lambda epoch: max(1 - epoch * factor, 0))
        return scheduler

    @staticmethod
    def create_agent(FLAGS, observation_space, action_space):
        model = Impala._create_model(FLAGS, observation_space, action_space)
        optimizer = __Optimizer__.build(FLAGS.model.optimizer, model.parameters())
        scheduler = Impala._create_scheduler(FLAGS, optimizer)
        learner_state = ImpalaLearnerState(model, optimizer, scheduler)
        return model, learner_state

    @staticmethod
    def create_stats(FLAGS):
        stats_keys = [
            "entropy_loss",
            "pg_loss",
            "baseline_loss",
            "kl_loss",
            "total_loss",
            "unclipped_grad_norm",
        ]

        stats = {}
        for k in stats_keys:
            stats[k] = common.StatMean()
        return stats

    @staticmethod
    def step_optimizer(FLAGS, learner_state, stats):
        MFLAGS = FLAGS.model
        unclipped_grad_norm = nn.utils.clip_grad_norm_(learner_state.model.parameters(), MFLAGS.grad_norm_clipping)
        learner_state.optimizer.step()
        learner_state.scheduler.step()
        learner_state.model_version += 1

        stats["unclipped_grad_norm"] += unclipped_grad_norm.item()
        stats["optimizer_steps"] += 1
        stats["model_version"] += 1
        return

    @staticmethod
    def compute_gradients(FLAGS, data, learner_state, stats):
        MFLAGS = FLAGS.model
        model = learner_state.model

        env_outputs = data["env_outputs"]
        actor_outputs = data["actor_outputs"]
        initial_core_state = data["initial_core_state"]

        model.train()

        learner_outputs, _ = model(env_outputs, initial_core_state)

        # Use last baseline value (from the value function) to bootstrap.
        bootstrap_value = learner_outputs["baseline"][-1]

        # Move from env_outputs[t] -> action[t] to action[t] -> env_outputs[t].
        # seed_rl comment: At this point, we have unroll length + 1 steps. The last step is only used
        # as bootstrap value, so it's removed.
        learner_outputs = nest.map(lambda t: t[:-1], learner_outputs)
        env_outputs = nest.map(lambda t: t[1:], env_outputs)
        actor_outputs = nest.map(lambda t: t[:-1], actor_outputs)

        rewards = env_outputs["reward"]
        if MFLAGS.reward_clip:
            rewards = torch.clip(rewards, -MFLAGS.reward_clip, MFLAGS.reward_clip)

        # TODO: reward normalization ?

        discounts = (~env_outputs["done"]).float() * MFLAGS.discounting

        # an attempt at supporting continuous action spaces based on
        # https://github.com/google-research/seed_rl/search?q=continuous&type=commits

        behavior_policy_action_dist = model.policy.action_dist(actor_outputs["policy_logits"])
        target_policy_action_dist = model.policy.action_dist(learner_outputs["policy_logits"])

        actions = actor_outputs["action"]
        behavior_action_log_probs = behavior_policy_action_dist.log_prob(actions)
        target_action_log_probs = target_policy_action_dist.log_prob(actions)

        log_rhos = target_action_log_probs - behavior_action_log_probs
        # TODO: put this on cpu? https://github.com/deepmind/scalable_agent/blob/6c0c8a701990fab9053fb338ede9c915c18fa2b1/experiment.py#L374
        # or move to C++ https://github.com/facebookresearch/minihack/blob/65fc16f0f321b00552ca37db8e5f850cbd369ae5/minihack/agent/polybeast/polybeast_learner.py#L342
        vtrace_returns = from_importance_weights(
            log_rhos=log_rhos,
            discounts=discounts,
            rewards=rewards,
            values=learner_outputs["baseline"],
            bootstrap_value=bootstrap_value,
            clip_rho_threshold=1.0,
            clip_pg_rho_threshold=1.0,
        )

        # TODO target entropy adjustment: https://github.com/google-research/seed_rl/blob/66e8890261f09d0355e8bf5f1c5e41968ca9f02b/agents/vtrace/learner.py#L127
        entropy_loss = MFLAGS.entropy_cost * -target_policy_action_dist.entropy().mean()

        log_likelihoods = target_action_log_probs  # target_policy_action_dist.log_prob(actions)
        pg_loss = -torch.mean(log_likelihoods * vtrace_returns.pg_advantages.detach())  # policy gradient

        baseline_advantages = vtrace_returns.vs - learner_outputs["baseline"]
        baseline_loss = MFLAGS.baseline_cost * (0.5 * torch.mean(baseline_advantages**2))

        # from https://github.com/google-research/seed_rl/blob/66e8890261f09d0355e8bf5f1c5e41968ca9f02b/agents/vtrace/learner.py#L123
        # KL(old_policy|new_policy) loss
        kl_loss = MFLAGS.kl_cost * torch.mean(behavior_action_log_probs - target_action_log_probs)
        # TODO: could also use `D.kl_divergence(behavior_policy_action_dist, target_policy_action_dist)`

        total_loss = entropy_loss + pg_loss + baseline_loss + kl_loss
        total_loss.backward()

        stats["entropy_loss"] += entropy_loss.item()
        stats["pg_loss"] += pg_loss.item()
        stats["baseline_loss"] += baseline_loss.item()
        stats["kl_loss"] += kl_loss.item()
        stats["total_loss"] += total_loss.item()
        return



import importlib
import os
import subprocess

from setuptools import Extension, find_packages, setup

from third_party.moolib.setup import CMakeBuild as MoolibCmakeBuild

NAME = "multibeast"
AUTHOR = f"{NAME} contributors"
URL = "https://github.com/etaoxing/multibeast"
__version__ = "0.0.1a1"

cwd = os.path.dirname(os.path.abspath(__file__))
packages = find_packages(exclude=["tests*", "examples*"]) + ["multibeast.examples"]
package_dir = {"": ".", "multibeast.examples": "examples"}
ext_modules = []
cmdclass = {}

install_requires = [
    "torch>=1.9.1",  # make sure this is the first entry
    "hydra-core>=1.0.0",
    "hydra-colorlog>=1.0.0",
    "hydra-submitit-launcher>=1.1.1",
    "wandb>=0.10.31",
    "pyyaml",
]

moolib_spec = importlib.util.find_spec("moolib")
found_moolib = moolib_spec is not None
if os.getenv("SKIP_MOOLIB_BUILD"):
    install_requires += [  # pypi version
        "moolib",
    ]
else:
    install_requires += [
        "cmake>=3.14.4",  # for building moolib
    ]
    if os.getenv("RELEASE_BUILD"):
        # bundle a build of moolib
        packages += ["moolib", "moolib.examples.common"]
        package_dir.update({"moolib": "third_party/moolib/py/moolib", "moolib.examples": "third_party/moolib/examples"})
        ext_modules += [Extension("moolib._C", sources=[])]
        cmdclass["build_ext"] = MoolibCmakeBuild

        # enforce a strict pytorch version
        install_requires[0] = "==".join(install_requires[0].split(">="))
    else:
        # support pip installing in editable mode
        # otherwise top-level moolib module will not be found
        install_requires += [
            f"moolib @ file://{cwd}/third_party/moolib#egg=moolib",
        ]

install_requires += [
    "tinyspace",
    "tabulate",
    "coolname",
]

extras_deps = {
    "tests": [
        "pre-commit>=2.0.1",
        # Reformat
        "black>=19.10b0",
        # Lint code
        "flake8>=3.7",
        # Find likely bugs
        "flake8-bugbear>=20.1",
        # Docstrings style,
        "flake8-docstrings>=1.6.0",
        # Run tests and coverage
        "pytest>=5.3",
        "pytest-benchmark>=3.1.0",
        "pytest-order>=1.0.1",
        "pytest-cov",
        "pytest-xdist",
        # Type check
        "pyright",
        # Sort imports
        "isort>=5.0",
    ],
    "docs": [
        "sphinx==4.4.0",
        "sphinx-autobuild",
        "myst-parser",
        # # Jupyter notebooks
        # "nbsphinx",
        # For spelling
        "sphinxcontrib-spelling",
        # Type hints support
        "sphinx-autodoc-typehints",
        # Extras
        "sphinx-design",
        "sphinx-copybutton",
        "sphinx-inline-tabs",
        "sphinxcontrib-trio",
        "sphinxext-opengraph",
        # Theme
        "furo",
    ],
}

extras_deps["all"] = [item for group in extras_deps.values() for item in group]


if __name__ == "__main__":
    with open("README.md") as f:
        long_description = f.read()
    sha = "unknown"
    version = __version__

    if os.getenv("RELEASE_BUILD") or (os.getenv("READTHEDOCS") and os.getenv("READTHEDOCS_VERSION_TYPE") == "tag"):
        sha = version

        if not os.getenv("SKIP_MOOLIB_BUILD"):
            # platform-specific
            cuda = None
            try:
                nvcc_ver = os.popen('nvcc --version | egrep -o "V[0-9]+.[0-9]+.[0-9]+" | cut -c2-').read().strip()
                cu_major, cu_minor, cu_patch = nvcc_ver.split(".")
                cuda = f"{cu_major}{cu_minor}"
            except Exception:
                pass

            if cuda is not None:
                version = f"{__version__}+cu{cuda}"
    else:
        try:
            sha = subprocess.check_output(["git", "rev-parse", "HEAD"], cwd=cwd).decode("ascii").strip()
        except subprocess.CalledProcessError:
            pass
        version += ".dev0+" + sha[:7]

    version_path = os.path.join(cwd, NAME, "version.py")
    with open(version_path, "w") as f:
        f.write(f'__version__ = "{version}"\n')
        f.write(f'commit = "{sha}"\n')

    print(f"Building {NAME}-{version}")
    print(packages)

    setup(
        name=NAME,
        version=version,
        description="",
        author="etaoxing",
        url=URL,
        download_url=f"{URL}/tags",
        license="MIT",
        packages=packages,
        package_dir=package_dir,
        ext_modules=ext_modules,
        cmdclass=cmdclass,
        include_package_data=True,
        install_requires=install_requires,
        extras_require=extras_deps,
        python_requires=">=3.7",
        zip_safe=False,
    )



# from https://github.com/SudeepDasari/one_shot_transformers/blob/ecd43b0c182451b67219fdbc7d6a3cd912395f17/hem/models/discrete_logistic.py#L1

# some code borrowed from r9y9's wavenet vocoder code
# https://github.com/r9y9/wavenet_vocoder/blob/master/wavenet_vocoder/mixture.py

import numpy as np
import torch
import torch.nn.functional as F
from torch.distributions import Distribution

from ..builder import __Distribution__


def to_one_hot(tensor, n, fill_with=1.0):
    # we perform one hot encore with respect to the last axis
    one_hot = torch.FloatTensor(tensor.size() + (n,)).zero_()
    if tensor.is_cuda:
        one_hot = one_hot.cuda()
    one_hot.scatter_(len(tensor.size()), tensor.unsqueeze(-1), fill_with)
    return one_hot


# TODO: clamp log_scale_max? like https://github.com/ikostrikov/jaxrl/blob/8ac614b0c5202acb7bb62cdb1b082b00f257b08c/jaxrl/networks/policies.py#L71
# https://github.com/r9y9/wavenet_vocoder/blob/a35fff76ea3687b05e1a10023cad3f7f64fa25a3/wavenet_vocoder/mixture.py
# see A.2 of https://arxiv.org/pdf/1903.01973.pdf
# and the original ref: https://arxiv.org/pdf/1701.05517.pdf
# https://www.tensorflow.org/probability/api_docs/python/tfp/distributions/PixelCNN
@__Distribution__.register()
class DiscretizedLogisticMixture(Distribution):
    def __init__(self, mean, log_scale, logit_probs, num_bins=256, log_scale_min=-7.0, validate_args=None):
        batch_shape = log_scale.shape[:-1]
        event_shape = mean.shape[len(batch_shape) + 1 :]
        super().__init__(batch_shape, event_shape, validate_args)
        self._mean = mean
        self._log_scale = log_scale
        self._logit_probs = logit_probs
        self._num_classes = num_bins
        self._log_scale_min = log_scale_min

    @property
    def arg_constraints(self):
        return {}

    def log_prob(self, value):
        # reshape value to match convention
        B, n_mix = value.shape[0], self._log_scale.shape[-1]

        # unpack parameters. (B, T, num_mixtures) x 3
        logit_probs = self._logit_probs.reshape((self._log_scale.shape[0], -1, n_mix))
        means = self._mean.reshape((self._mean.shape[0], -1, n_mix))
        log_scales = torch.clamp(
            self._log_scale.reshape((self._log_scale.shape[0], -1, n_mix)),
            min=self._log_scale_min,
        )

        # B x T x 1 -> B x T x num_mixtures
        y = value.reshape((B, -1, 1))

        centered_y = y - means
        inv_stdv = torch.exp(-log_scales)
        plus_in = inv_stdv * (centered_y + 1.0 / (self._num_classes - 1))
        cdf_plus = torch.sigmoid(plus_in)
        min_in = inv_stdv * (centered_y - 1.0 / (self._num_classes - 1))
        cdf_min = torch.sigmoid(min_in)

        # log probability for edge case of 0 (before scaling)
        # equivalent: torch.log(torch.sigmoid(plus_in))
        log_cdf_plus = plus_in - F.softplus(plus_in)

        # log probability for edge case of 255 (before scaling)
        # equivalent: (1 - torch.sigmoid(min_in)).log()
        log_one_minus_cdf_min = -F.softplus(min_in)

        # probability for all other cases
        cdf_delta = cdf_plus - cdf_min

        mid_in = inv_stdv * centered_y
        # log probability in the center of the bin, to be used in extreme cases
        # (not actually used in our code)
        log_pdf_mid = mid_in - log_scales - 2.0 * F.softplus(mid_in)

        # tf equivalent
        """
        log_probs = tf.where(x < -0.999, log_cdf_plus,
                            tf.where(x > 0.999, log_one_minus_cdf_min,
                                    tf.where(cdf_delta > 1e-5,
                                            tf.log(tf.maximum(cdf_delta, 1e-12)),
                                            log_pdf_mid - np.log(127.5))))
        """
        # TODO: cdf_delta <= 1e-5 actually can happen. How can we choose the value
        # for num_classes=65536 case? 1e-7? not sure..
        inner_inner_cond = (cdf_delta > 1e-5).float()

        inner_inner_out = inner_inner_cond * torch.log(torch.clamp(cdf_delta, min=1e-12)) + (1.0 - inner_inner_cond) * (
            log_pdf_mid - np.log((self._num_classes - 1) / 2)
        )
        inner_cond = (y > 0.999).float()
        inner_out = inner_cond * log_one_minus_cdf_min + (1.0 - inner_cond) * inner_inner_out
        cond = (y < -0.999).float()
        log_probs = cond * log_cdf_plus + (1.0 - cond) * inner_out

        log_probs = log_probs + F.log_softmax(logit_probs, -1)
        return torch.logsumexp(log_probs, axis=-1).reshape(value.shape)

    def rsample(self):
        n_mix = self._log_scale.shape[-1]
        logit_probs = self._logit_probs.reshape((self._log_scale.shape[0], -1, n_mix))

        # sample mixture indicator from softmax
        temp = logit_probs.data.new(logit_probs.size()).uniform_(1e-5, 1.0 - 1e-5)
        temp = logit_probs.data - torch.log(-torch.log(temp))
        _, argmax = temp.max(dim=-1)

        # (B, T) -> (B, T, nr_mix)
        one_hot = to_one_hot(argmax, n_mix)
        # select logistic parameters
        means = self._mean.reshape((self._mean.shape[0], -1, n_mix))
        means = torch.sum(means * one_hot, dim=-1)
        log_scales = self._log_scale.reshape((self._log_scale.shape[0], -1, n_mix))
        log_scales = torch.sum(log_scales * one_hot, dim=-1)

        # sample from logistic & clip to interval
        # we don't actually round to the nearest 8bit value when sampling
        u = means.data.new(means.size()).uniform_(1e-5, 1.0 - 1e-5)
        x = means + torch.exp(log_scales) * (torch.log(u) - torch.log(1.0 - u))

        x = torch.clamp(torch.clamp(x, min=-1.0), max=1.0)

        return x.reshape(self._mean.shape[:-1])

    @property
    def mean(self):
        alphas = F.softmax(self._logit_probs, dim=-1)
        return torch.sum(self._mean * alphas, -1)



import logging
import os
import pprint

import moolib
import omegaconf
import torch
from moolib.examples.common import record


def log(stats, step, is_global=False, wandb=False):
    stats_values = {}
    prefix = "global/" if is_global else "local/"
    for k, v in stats.items():
        stats_values[prefix + k] = v.result()
        v.reset()

    logging.info(f"\n{pprint.pformat(stats_values)}")
    if not is_global:
        record.log_to_file(**stats_values)

    if wandb:
        import wandb

        wandb.log(stats_values, step=step)


def save_checkpoint(checkpoint_path, learner_state, flags):
    tmp_path = "%s.tmp.%s" % (checkpoint_path, moolib.create_uid())

    logging.info("saving global stats %s", learner_state.global_stats)

    checkpoint = {
        "learner_state": learner_state.save(),
        "flags": omegaconf.OmegaConf.to_container(flags),
    }

    torch.save(checkpoint, tmp_path)
    os.replace(tmp_path, checkpoint_path)

    logging.info("Checkpoint saved to %s", checkpoint_path)


def load_checkpoint(checkpoint_path, learner_state):
    checkpoint = torch.load(checkpoint_path)
    learner_state.load(checkpoint["learner_state"])


def calculate_sps(stats, delta, prev_steps, is_global=False):
    env_train_steps = stats["env_train_steps"].result()
    prefix = "global/" if is_global else "local/"
    logging.info("%s calculate_sps %g steps in %g seconds", prefix, env_train_steps - prev_steps, delta)
    stats["SPS"] += (env_train_steps - prev_steps) / delta
    return env_train_steps



# Copyright (c) Facebook, Inc. and its affiliates.
from ._C import (
    Accumulator,
    AllReduce,
    Batcher,
    Broker,
    EnvPool,
    EnvRunner,
    EnvStepper,
    EnvStepperFuture,
    Future,
    Group,
    Queue,
    Rpc,
    RpcDeferredReturn,
    RpcError,
    __doc__,
    create_uid,
    set_log_level,
    set_logging,
    set_max_threads,
)


__all__ = [
    "Accumulator",
    "AllReduce",
    "Batcher",
    "Broker",
    "EnvPool",
    "EnvRunner",
    "EnvStepper",
    "EnvStepperFuture",
    "Future",
    "Group",
    "Queue",
    "Rpc",
    "RpcDeferredReturn",
    "RpcError",
    "__doc__",
    "create_uid",
    "set_log_level",
    "set_logging",
    "set_max_threads",
]



import torch
import torch.distributions as D

from ..builder import __Distribution__


class TanhBijector(object):
    """Bijective transformation of a probability distribution using a squashing function (tanh).

    TODO: use Pyro instead (https://pyro.ai/)
    """

    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(x: torch.Tensor) -> torch.Tensor:
        return torch.tanh(x)

    @staticmethod
    def atanh(x: torch.Tensor) -> torch.Tensor:
        """Inverse of Tanh.

        Taken from Pyro: https://github.com/pyro-ppl/pyro
        0.5 * torch.log((1 + x ) / (1 - x))
        """
        return 0.5 * (x.log1p() - (-x).log1p())

    @staticmethod
    def inverse(y: torch.Tensor) -> torch.Tensor:
        eps = torch.finfo(y.dtype).eps
        # Clip the action to avoid NaN
        return TanhBijector.atanh(y.clamp(min=-1.0 + eps, max=1.0 - eps))


@__Distribution__.register()
class SquashedDiagGaussian(D.Distribution):
    r"""Also called a TanhNormal distribution.

    from https://github.com/DLR-RM/stable-baselines3/blob/ed308a71be24036744b5ad4af61b083e4fbdf83c/stable_baselines3/common/distributions.py#L195

    Also see:
    - https://github.com/rlworkgroup/garage/blob/c56513f42be9cba2ef5426425a8ad36097e679c2/src/garage/torch/distributions/tanh_normal.py
    - https://github.com/ikostrikov/jaxrl/blob/5e044e1fd080223ef5e71848b9d0f72597f2ab2f/jax_rl/networks/policies.py#L15
    """

    arg_constraints = D.Normal.arg_constraints
    support = D.Normal.support
    has_rsample = True

    def __init__(self, loc, scale, eps: float = 1e-6, validate_args=None):
        batch_shape = loc.shape[:-1]
        event_shape = loc.shape[len(batch_shape) + 1 :]
        self.loc = loc
        self.scale = scale
        super().__init__(batch_shape, event_shape, validate_args)
        self.gaussian = D.Independent(D.Normal(loc, scale), 1)
        # Avoid NaN (prevents division by zero or log of zero)
        self.eps = eps

    def _squash_correction(self, value):
        # see https://github.com/haarnoja/sac/blob/8258e33633c7e37833cc39315891e77adfbe14b2/sac/policies/gaussian_policy.py#L131
        # Squash correction (from original SAC implementation)
        # this comes from the fact that tanh is bijective and differentiable
        return torch.sum(torch.log(1 - value**2 + self.eps), dim=1)

    def log_prob(self, value):
        # Inverse tanh
        # Naive implementation (not stable): 0.5 * torch.log((1 + x) / (1 - x))
        # We use numpy to avoid numerical instability

        # It will be clipped to avoid NaN when inversing tanh
        gaussian_value = TanhBijector.inverse(value)

        # Log likelihood for a Gaussian distribution
        log_prob = self.gaussian.log_prob(gaussian_value)
        # don't need https://github.com/DLR-RM/stable-baselines3/blob/ed308a71be24036744b5ad4af61b083e4fbdf83c/stable_baselines3/common/distributions.py#L100
        # since we use D.Independent

        if len(value.shape) == 3:  # deal with (T, B, ...) shaped inputs
            T, B = log_prob.shape
            log_prob = log_prob.view(T * B) - self._squash_correction(value.view(T * B, -1))
            log_prob = log_prob.view(T, B)
        else:
            log_prob -= self._squash_correction(value)
        return log_prob

    def rsample(self, sample_shape=torch.Size()):
        sampled = self.gaussian.rsample(sample_shape=sample_shape)
        return torch.tanh(sampled)

    def mode(self):
        r"""Deterministically returns the most likely sample from the distribution."""
        return torch.tanh(self.gaussian.mean)

    def entropy(self):
        # No analytical form,
        # entropy needs to be estimated using -log_prob.mean()
        # return None

        # just return the underlying gaussian's entropy
        return self.gaussian.entropy()



import torch
import torch.distributions as D

from ..builder import __Distribution__


@__Distribution__.register()
class NormalMixture(D.Distribution):
    def __init__(self, mean, log_scale, logit_probs, temperature=1.0):
        batch_shape = log_scale.shape[:-1]
        event_shape = mean.shape[len(batch_shape) + 1 :]
        super().__init__(batch_shape, event_shape, None)

        mix_ = D.Categorical(logits=logit_probs)
        comp = D.Independent(D.Normal(mean, torch.exp(log_scale) * temperature), 1)
        dist = D.MixtureSameFamily(mix_, comp)

        self.dist = dist
        # see https://github.com/ikostrikov/jaxrl/blob/8ac614b0c5202acb7bb62cdb1b082b00f257b08c/jaxrl/networks/policies.py#L47
        # https://pytorch.org/docs/stable/distributions.html#mixturesamefamily

        # TODO: sigmoid or tanh when gaussianpolicy?
        # https://github.com/rail-berkeley/rlkit/blob/354f14c707cc4eb7ed876215dd6235c6b30a2e2b/rlkit/torch/sac/policies/gaussian_policy.py#L132

    def log_prob(self, value):
        log_probs = self.dist.log_prob(value)
        assert log_probs.shape == value.shape
        return log_probs

    def sample(self):
        return self.dist.sample()

    @property
    def mean(self):
        return self.dist.mean



from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F
from moolib.examples.common import nest
from tinyspace import Space, sample_from_space

from multibeast.builder import __PolicyNet__
from multibeast.distributions import get_action_params


class PPONet(nn.Module):
    def __init__(
        self,
        observation_space: Space,
        action_space: Space,
        action_dist_params: Dict = None,
        policy_params: Dict = None,
        feature_extractor: nn.Module = None,
        use_lstm: bool = False,
        fc_out_features: int = 256,
    ):
        super().__init__()
        self.observation_space = observation_space
        self.action_space = action_space

        self.num_actions, self.max_action, self.action_dist_params = get_action_params(
            action_space,
            action_dist_params,
        )

        self.feature_extractor = feature_extractor
        self.use_lstm = use_lstm

        if feature_extractor:
            with torch.no_grad():
                x = sample_from_space(self.observation_space, batch_size=(1, 1), to_torch_tensor=True)
                feature_dim = tuple(self.feature_extractor(x).squeeze(0).shape)[0]
            self.fc = nn.Sequential(nn.ReLU(inplace=True), nn.Linear(feature_dim, fc_out_features), nn.ReLU(inplace=True))
        else:
            feature_dim = observation_space["shape"][0]
            self.fc = nn.Sequential(nn.Linear(feature_dim, fc_out_features), nn.ReLU(inplace=True))

        # FC output size + last action (one-hot if discrete) + last reward.
        core_output_size = fc_out_features + self.num_actions + 1

        if use_lstm:
            self.core = nn.LSTM(core_output_size, 256, num_layers=1)
            core_output_size = 256

        if policy_params is None:
            policy_params = dict(cls="PolicyNet")
        self.policy = __PolicyNet__.build(policy_params, core_output_size, self.num_actions, self.action_dist_params)
        self.baseline = nn.Linear(core_output_size, 1)

    def initial_state(self, batch_size=1):
        if self.use_lstm:
            return tuple(torch.zeros(self.core.num_layers, batch_size, self.core.hidden_size) for _ in range(2))
        return tuple()

    def forward(self, inputs, core_state=None, deterministic: bool = False):
        prev_action = inputs["prev_action"]
        reward = inputs["reward"]
        x = inputs["state"]

        T, B = reward.shape
        if self.feature_extractor:
            x = self.feature_extractor(x)
            if len(x.shape) != 2:
                raise RuntimeError(f"Expected feature extractor to output batch of 1d features, got {x.shape}")

        x = x.view(T * B, -1)
        x = self.fc(x)

        clipped_reward = torch.clamp(reward, -1, 1).view(T * B, 1)
        if self.action_space["cls"] == "discrete":
            if prev_action.dtype != torch.long:
                # if getting "RuntimeError: one_hot is only applicable to index tensor."
                # then prev_action is not a long
                raise RuntimeError
            last_action = F.one_hot(prev_action.view(T * B), self.num_actions).float()
        elif self.action_space["cls"] == "box":
            last_action = prev_action.view(T * B, -1)
        else:
            raise ValueError
        core_input = torch.cat([x, clipped_reward, last_action], dim=-1)

        if self.use_lstm:
            done = inputs["done"]
            core_input = core_input.view(T, B, -1)
            core_output_list = []
            notdone = (~done).float()
            for input, nd in zip(core_input.unbind(), notdone.unbind()):
                # Reset core state to zero whenever an episode ended.
                # Make `done` broadcastable with (num_layers, B, hidden_size)
                # states:
                nd = nd.view(1, -1, 1)
                core_state = nest.map(nd.mul, core_state)
                core_output, core_state = self.core(input.unsqueeze(0), core_state)
                core_output_list.append(core_output)
            core_output = torch.flatten(torch.cat(core_output_list), 0, 1)
        else:
            core_output = core_input

        core_output = core_output.view(T, B, -1)
        policy_logits, action = self.policy(core_output, deterministic=deterministic)
        if self.action_space["cls"] == "discrete":
            action = action.view(T, B)

        baseline = self.baseline(core_output)
        baseline = baseline.view(T, B)

        output = dict(
            policy_logits=policy_logits,
            action=action,
            baseline=baseline,
        )
        return output, core_state



# Copyright (c) Facebook, Inc. and its affiliates.


def map(f, n):
    if isinstance(n, tuple) or isinstance(n, list):
        return n.__class__(map(f, sn) for sn in n)
    elif isinstance(n, dict):
        return {k: map(f, v) for k, v in n.items()}
    else:
        return f(n)


def flatten(n):
    if isinstance(n, tuple) or isinstance(n, list):
        for sn in n:
            yield from flatten(sn)
    elif isinstance(n, dict):
        for key in n:
            yield from flatten(n[key])
    else:
        yield n


def zip(*nests):
    n0, *nests = nests
    iters = [flatten(n) for n in nests]

    def f(first):
        return [first] + [next(i) for i in iters]

    return map(f, n0)


def map_many(f, *nests):
    n0, *nests = nests
    iters = [flatten(n) for n in nests]

    def g(first):
        return f([first] + [next(i) for i in iters])

    return map(g, n0)



# based on https://github.com/facebookresearch/moolib/blob/e8b2de7ac5df3a9b3ee2548a33f61100a95152ef/examples/common/__init__.py#L154

import asyncio
from typing import Any, Callable

import moolib
import torch

from .envstepper import EnvStepper


class EnvPool:
    r"""A wrapper class to manage a pool of environments.

    The class maintains `num_batches` batches of environments, with `batch_size` environments each.
    This means the batches can be stepped through alternately, for increased efficiency
    (cf "double-buffering"), and the whole `EnvPool` uses ``num_process` to run these environments.

    We use this instead of `moolib@b785684.EnvPool` b/c moolib's version only supports environments with
    discrete action spaces. Furthermore, moolib's version uses RPC and C++ threads, while this class is
    entirely in Python for simplicity. This class supports GPU-based environments (like IsaacGym) which
    are already batched, and provides a way to batch standard Gym-like environments using multiple processes.

    Args:
        create_env (Callable[[], Env]): a user-defined function that returns a Gym-like environment.
        num_processes: how many processes should be used for running environments.
        batch_size: the number of environments in one batch.
        num_batches: the number of batches to maintain (for double-buffering).
    """

    def __init__(
        self,
        env_init_fn: Callable[[], Any],
        num_processes: int,
        batch_size: int,
        num_batches: int,
    ):
        assert num_processes / batch_size == num_batches

        self.num_processes = num_processes
        self.batch_size = batch_size
        self.num_batches = num_batches

        self.env_stepper = self.spawn(env_init_fn)

    def spawn(self, env_init_fn):
        self.name = moolib.create_uid()

        EnvStepperCls = EnvStepper  # for unbatched, CPU-based environments (ie. Gym)
        # EnvStepperCls = BatchedEnvStepper  # use as a wrapper for batched, CPU-based environments (ie. sail-sg/envpool)
        # EnvStepperCls = BatchedCUDAEnvStepper  # for batched, GPU-based environments that use CUDA tensors (ie. IsaacGym)

        env_stepper = EnvStepperCls(env_init_fn, self.num_processes)
        return env_stepper

    def step(self, batch_index: int, action: torch.Tensor) -> "asyncio.Future":
        r"""Step through a batch of envs.

        Args:
            batch_index: index of the batch of envs are we stepping.
            action: actions for each of the envs.
        """
        if action.size(0) != self.batch_size:
            raise RuntimeError(
                f"env step was passed an action tensor with batch size {action.size(0)}, expected {self.batch_size}"
            )

        if batch_index < 0 or batch_index >= self.num_batches:
            raise RuntimeError(
                f"env step was passed an out-of-range batch index {batch_index} (valid range is [0,{self.num_batches})"
            )

        return self.env_stepper.step(batch_index, action)



def get_action_params(action_space, action_dist_params):
    if action_space["cls"] == "discrete":
        num_actions = action_space["high"]
        max_action = num_actions
        if action_dist_params is None:
            action_dist_params = dict(cls="Categorical")  # same as `torch.multinomial`

    elif action_space["cls"] == "box":
        num_actions = action_space["shape"][0]
        try:
            max_action = max(action_space["high"])
        except:  # TypeError: 'float' object is not iterable
            max_action = action_space["high"]
        if max_action != 1.0:
            raise RuntimeError("Currently only supporting continuous action spaces bound to [-1.0, 1.0]")

        if action_dist_params is None:
            action_dist_params = dict(cls="SquashedDiagGaussian")
    else:
        raise ValueError

    return num_actions, max_action, action_dist_params
